/* ---------------------------------------------------------------------------
 * IBM VGA chipset driver
 * --------------------------------------------------------------------------
 *
 * Copyright (C) 1996, 1997 The Linux-GGI Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * ---------------------------------------------------------------------------
 * 
 * $Log: vga.c,v $
 * Revision 1.43  1997/12/18 00:19:33  becka
 * Killed a few bugs.
 *
 * Revision 1.42  1997/12/14 21:55:37  becka
 * uint/sint namespace fixed ... please test ...
 *
 * Revision 1.41  1997/12/13 19:32:19  becka
 * VGA fixes for ->private. autoconfig more PCI detection.
 *
 * Revision 1.40  1997/12/09 16:27:43  becka
 * VGA driver fixed.
 *
 * Revision 1.39  1997/12/09 10:11:37  seeger
 * Global cleanup for new include directory locations.
 *
 * Revision 1.38  1997/11/29 18:01:00  taylorj
 * Major: Committed Brian Julin's patch for chipset_set_mode.  It now uses the
 * header macro values and doesn;t rely on the chipset default mode being
 * blanked out in cartain areas.  Also fixed a small bug in the `set splitline
 * to zero' fragment.
 *
 * Revision 1.37  1997/11/08 15:12:26  becka
 * Namespace cleanup ... probably broke all and everything ...
 *
 * Revision 1.36  1997/11/03 21:22:13  taylorj
 * Minor: Fixed a couple of missing typecasts.
 *
 * Revision 1.35  1997/10/29 20:21:29  becka
 * New driver - aalib
 * VGA fixes
 *
 * Revision 1.34  1997/10/19 17:58:52  becka
 * Updated docs for libggi
 * Fixed VGA driver for new timings.
 * Added new card.
 * Minor demo additions
 *
 * Revision 1.33  1997/09/12 07:28:41  taylorj
 * Minor: Reformatted source to be consistent with GGI coding style guidelines,
 * modified all debugging output to use the DEBUGx macros, and added CVS
 * revision logging to headers where needed.
 *
 * Revision 1.32  1997/09/06 04:06:42  taylorj
 * *** empty log message ***
 *
 * Revision 1.31  1997/08/23 03:42:09  taylorj
 * Preliminary CHIP_SETDEFFONT ioctl support
 *
 * Revision 1.30  1997/08/04 23:46:49  taylorj
 * Added virtual base address field to io_region struct decl to match the new
 * def. in include/system.h
 *
 * Revision 1.29  1997/07/01 04:29:46  ezrec
 * Minor: Alpha attempt at ggiSetSplitline support
 *
 * Revision 1.28  1997/06/28 18:49:17  becka
 * vga endsync bug fixed
 *
 * Revision 1.27  1997/06/03 04:47:38  ezrec
 * Minor: Fixed VGA `splitline' bug (I wasn't zeroing-out the fb.ops for
 * graphics modes), and a vga_modex drawing bug.
 *
 * Revision 1.26  1997/05/12 14:38:22  ezrec
 * Minor: fixed splitline problem in NNNx480 modes
 *
 * Revision 1.25  1997/04/28 04:31:51  ezrec
 * Minor: Fixed GT_1BIT support
 *
 * Revision 1.24  1997/04/27 17:27:17  ezrec
 * Minor: vga.c compiles again.  Still need to test it...
 *
 * Revision 1.23  1997/04/23 03:52:44  ezrec
 * Minor: Working on the ATI Mach64...
 *
 * Revision 1.22  1997/04/17 05:08:03  ezrec
 * Major: New ggiOpen() semantics ie ggiOpen("display-KGI:/dev/graph3",NULL)
 * First attempt at linear_1 frame buffer
 * Moved thread support `to the side'
 * Added DPRINT() support
 * Added display-KGI driver
 *
 * Revision 1.21  1997/04/12 04:51:27  ezrec
 * Minor: Fixes to the VGA subsystem to work w/Steffen's new MMAP system
 *
 * Revision 1.20  1997/04/10 06:36:36  ezrec
 * Major: Fixes for Steffen's new MMAP...
 *
 * Revision 1.19  1997/04/10 03:34:01  ezrec
 * Major: More fixes, and I'm attempting (ha!) to use Steffen's new MMAP scheme..
 *
 * Revision 1.18  1997/04/09 04:17:59  ezrec
 * Major: Added palettes in include/, fixed 8bit for VGA, and
 * got libggi-dynamic operational... Now to make it work
 * for the _new_ MMAP scheme!
 *
 * Revision 1.17  1997/04/04 02:58:26  ezrec
 * Minor: More mmap stuff, with lots of kern.debug class messages...
 *
 * Revision 1.16  1997/03/27 14:28:54  becka
 * VGA-16 works. Sysinfo added.
 *
 * Revision 1.15  1997/03/26 23:50:58  becka
 * Better MMAP consistency checking. VGA mmap handler fixed. Still need
 * correct setpage.
 *
 * Revision 1.14  1997/03/22 00:05:15  ezrec
 * Minor: Fixed depend bug with pointer.h, and added more graphics support.
 * Still marginal and untested.
 *
 * Revision 1.13  1997/03/21 01:10:15  ezrec
 * Minor: First attempt at graphics support...
 *
 * Revision 1.12  1997/02/26 05:28:40  ezrec
 * Minor: Thanks to Massimiliano Ghilardi <ghilardi@cibs.sns.it>, the
 * font-cursor works very well... SLOW_POINTER is defined,
 * as I need it, but try without SLOW_POINTER and mail me or
 * M.G. the results...
 *
 * Revision 1.11  1997/02/21 13:57:55  ezrec
 * Minor: Added 'SLOW_POINTER' option, for people who don't like waiting
 * for vertical retrace... Default is undefined.
 *
 * Revision 1.10  1997/02/20 21:59:06  ezrec
 * Minor: jmcc fixes: keypress_wait bug on 2.1.x (console.c)
 * Jochen Hoenicke <Jochen.Hoenicke@Informatik.Uni-Oldenburg.DE> fixed:
 * IBM/vga drivers
 * No-display-on-driver-load bug (woo woo!)
 * set_origin/Attribute handling bugs...
 * Added TEXT_A000_64, for those of you with MDA or
 * Hercules cards (it's now the default, anyway)
 *
 * Revision 1.9  1997/02/20 00:35:17  ezrec
 * Minor: Fixed font-mouse bug in vga.c, bad-string bug in ps2aux.c,
 * 2.0.x/2.1.x verify_area and include fixes, added 2.1.25,2.1.26
 * patchability.
 *
 * Revision 1.8  1997/02/09 19:10:46  ezrec
 * Minor: Fixed up vga_text16.inc to use the `charcell' data type instead of
 * unsigned shorts, and set the default cursor to block (and a yellow
 * block, at that!)
 *
 * Revision 1.7  1997/02/08 23:14:03  ezrec
 * Minor: Fixed 'thinko' that led to Oopses under 2.1.25, and now I do
 * pointer hiding/unhiding more intelligently..  (Tin was having a
 * HELLUVA time w/the pointer before, but it's _real_ nice now...)
 *
 * Revision 1.6  1997/02/08 08:58:55  ezrec
 * Major: Set block cursor as default, and added font-hack pointer.  Modify
 * pointer.h to suit your taste.... Still 'wiggles' a lot, but better
 * than it used to be...
 *
 * Revision 1.5  1997/02/06 03:33:38  ezrec
 * Minor: Moahaha! set_origin now works properly, so we have _really_
 * fast A0000-based text modes (128K of mind-numbing text
 * scrolling power!) Undef 'TEXT_A000' for B8000 (32k) text
 * modes, and Undef 'CURSOR_HW' to get a solid cursor.
 * (for laptops...)
 *
 * Revision 1.4  1997/02/05 22:55:50  ezrec
 * Major: Fixed the 'cursor follows' point problem, and hardware/software
 * cursor is selectable via the 'CURSOR_HW' define in vga.c
 * NOTE: You _must_ change video modes before anything will display..
 *
 * Revision 1.3  1997/02/04 04:45:48  ezrec
 * Minor: Turned off underline in text modes (should be IOCTL settable..)
 *
 * Revision 1.2  1997/02/04 04:33:44  ezrec
 * Minor: Small bugfixes to cursor and pointer code, but not fixed yet....
 *
 * Revision 1.1  1997/02/03 15:06:34  ezrec
 * Major: IBM vga driver (initial version, taken from GGI 0.0.8)
 *
 * ---------------------------------------------------------------------------
 * 
 * This code was derived from the following sources of information:
 * [1] Xfree86 VGA driver
 * [2] Zen of Graphics Programming (Michael Abrash)
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define DRIVER_NAME	"IBM VGA chipset driver"
#define DRIVER_REV	"$Revision: 1.43 $"

#define DEBUG_LEVEL	0

#include <ggi/module.h>
#include "IBM/vga.h"
#include "vga_io.inc"

#undef 	CURSOR_HW	/* Hardware cursor */
#define TEXT_A000	/* Base of textmode framebuffer */
#define	POINTER_FONT	/* Font-hack mouse cursor in textmodes */
#define SLOW_POINTER	/* #define this if your pointer flickers */

static struct kgi_chipset chipinfo =
{
	"IBM",			/* Manufacturer				*/
	"VGA",			/* Model				*/
	GM_TEXT16 | GM_1BIT |	/* Available modes			*/
	GM_4BIT | GM_8BIT,
	0,			/* Flags				*/
	{ 920, 550 },		/* Maximum resolution			*/
	0x40000,		/* 256k (quasi-linear) addressable RAM	*/
	{ 25000000, 32000000 }	/* DCLK limits				*/
};

/* This table holds the default values for the VGA registers.  A copy	*/
/* of it will be filled in as appropriate during mode programming in	*/
/* kgim_chipset_set_mode().						*/
static struct statetable default_state =
{
	0x23,							/* MISC	*/
	{
		0x03, 0x00, 0x00, 0x00, 0x00			/* SEQ	*/
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* CRT	*/
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* 0x1x	*/
		0x00
	},
	{
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC	*/
		0xFF
	},
	{
		0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC	*/
		0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
		0x20, 0x00, 0x0F, 0x00, 0x00
	}
};

struct statetable reset_state;
ggi_mode default_ggi_mode;
struct kgi_mode default_kgi_mode;

static unsigned long fb_text_len;
static unsigned long fb_text;
static unsigned long fb_graph_len;
static unsigned long fb_graph;

static struct io_region vga_io = 
{
	IO_NULL,		/* IO region start (virtual)	*/
  	0x3C0,			/* IO region start (physical)	*/
	0x20,			/* IO region size		*/
	0x3FF,			/* 10 address bits are decoded	*/
	"IBM VGA IO"		/* IO region name		*/
};

static struct mem_region vga_mem_graph = 
{
	MEM_NULL,		/* Mem region start (virtual)	*/
	0xA0000,		/* Mem region start (physical)	*/
	0x10000,		/* Graphics (64K) region	*/
	-1,			/* Decode all address bits	*/
	"IBM VGA - graphics"	/* Mem region name		*/
};

#ifndef TEXT_A000
static struct mem_region vga_mem_text = 
{
	MEM_NULL,		/* Mem region start (virtual)	*/
	0xB8000,		/* Mem region start (physical)	*/
	0x08000,		/* Text (32K) region		*/
	-1,			/* Decode all address bits	*/
	"IBM VGA - textmode"	/* Mem region name		*/
};
#endif

static int detect_VGA(void)
{
	int err;

	/* BUG: We should detect the VGA instead of assuming its presence */
	if ((err = io_check_region(&vga_io)))
		return err;
	
	outbMISC(inb(MISCr));	/* Adjust IO base       */

#ifdef TEXT_A000
	fb_text		= 0xA0000;
	fb_text_len	= 0x10000;	/* 64k framebuffer */
#else
	fb_text		= 0xB8000;
	fb_text_len	= 0x8000;
#endif /* TEXT_A000 */

	fb_graph_len = 0x20000;	/* We have only 64K linear graphics */
	fb_graph = fb_text;

	return EOK;
}

/* GGI interface */

/* Mode specific frame buffers */

#ifdef POINTER_FONT
# include "vga_text16_enh.inc"
#else
# include "vga_text16.inc"
#endif

#include "vga_fb.inc"

#ifdef DEBUG_FB
# include "vga_debug_fb.inc"
#endif

static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
		case GT_TEXT16:
		dpy->set_font		= text16_set_font;
		dpy->set_origin		= text16_set_origin;
		dpy->set_split		= text16_set_split;
		dpy->set_frame		= text16_set_frame;
		dpy->show_cursor	= text16_show_cursor;
		dpy->hide_cursor	= text16_hide_cursor;
		dpy->undo_cursor	= text16_undo_cursor;
		dpy->show_pointer	= text16_show_pointer;
		dpy->hide_pointer	= text16_hide_pointer;
		dpy->undo_pointer	= text16_undo_pointer;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x 		= mode->request.virt.x / 
			mode->request.text.x;
		dpy->fb.size.y		= mode->request.virt.y / 
			mode->request.text.y;
		dpy->fb.bpp		= MAP_16_BPP;
		dpy->fb.pbuf		= phys_to_virt(fb_text);
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= &text16_color_io;
		KGIM_PRIVATE(dpy, driver) = text16_data_realloc(KGIM_PRIVATE(dpy,driver));
		if (KGIM_PRIVATE(dpy, driver) == NULL)
			INTERNALERROR;
		break;
		default:
		dpy->set_font		= NULL;
		dpy->set_origin		= vgafb_set_origin;
		dpy->set_split		= vgafb_set_split;
		dpy->set_frame		= vgafb_set_frame;
		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= 0;
		dpy->fb.size.y		= 0;
		dpy->fb.bpp		= 0;
		dpy->fb.pbuf		= NULL;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= NULL;
		kgim_graphics_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode), mode, sizeof(struct kgi_mode));
}

/* Takes a mode and a bits-per-pixel value */
static void chipset_setup_mode(struct kgi_mode *mode)
{
	int bpp=0;

	/* By default, setup for graphics modes */

	switch (mode->request.graphtype) {
		case GT_TEXT16:
		mode->fb	= (mem_vaddr)fb_text;
		mode->fb_size	= fb_text_len;
		mode->fsize	= (mode->request.virt.x / mode->request.text.x) *
			(mode->request.virt.y / mode->request.text.y) * 2;
		mode->mem	= &text16_color_mem;
		mode->io	= &text16_color_io;
		return; 	/* Done setting up */
		case GT_1BIT:
		bpp = 1;
		break;
		case GT_4BIT:
		bpp = 4;
		break;
		case GT_8BIT:
		bpp = 8;
		break;
		default:
		INTERNALERROR;
		return;
	}
	
	mode->fb	= (mem_vaddr)fb_graph;
	mode->fb_size	= fb_graph_len;
	mode->fsize	= (mode->request.virt.x * 
			   mode->request.virt.y * 8) / bpp;
	mode->mem	= NULL;
	mode->io	= NULL;
}

static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	ggi_graphtype gmode = GT_TEXT16;
	int magx = 1, magy = 1;
	int DCLKdiv = 8;
	int dotclock;
	struct kgi_timing tx,ty;

	save_state(&reset_state);

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR5_WRITEMODE3) {
			case GR5_WRITEMODE0:
			gmode = GT_8BIT; /* 8 bit mode */
			break;
			case GR5_WRITEMODE1:
			gmode = GT_1BIT; /* 1 bit mode */
			break;
			case GR5_WRITEMODE3:
			gmode = GT_4BIT; /* 4 bit mode */
			break;
			case GR5_WRITEMODE2:
			break;
		}
		magy = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	} else { /* Text mode */
		gmode = GT_TEXT16;
		DCLKdiv = (reset_state.SEQ[0x01] & SR01_8DOT_CHCLK) ? 8 : 9;
	}
    
	/* Get horizontal timing */
	tx.width	= (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart	= reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart	= reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend	= ((reset_state.CRT[0x05]) & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend	= ((reset_state.CRT[0x03] & CR03_HBE_MASK) |
			   ((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
			   (reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total	= (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity	= (reset_state.MISC & MISC_NEG_HSYNC) != 0;

	/* Get vertical timing */
	ty.width	= 1 + (reset_state.CRT[0x12] | 
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart	= 1 + (reset_state.CRT[0x15] |
			       ((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
			       ((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart	= (reset_state.CRT[0x10] |                
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
			   ((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend	= (reset_state.CRT[0x11] & CR11_VSYNCEND_MASK) |
                (reset_state.CRT[0x12] & ~CR11_VSYNCEND_MASK);
	while (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend	= (reset_state.CRT[0x16] | ((ty.width - 1) & ~0xFF)) + 1;
	ty.total = 2 + (reset_state.CRT[0x06] |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
			((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC) != 0;
	
	dotclock = (getCLK() == MISC_25MHZ_CLK) ? 25175000 : 28322000;
   
	/* Set all the mode info that we've just gathered */
	default_ggi_mode.frames		= 1;
	default_ggi_mode.visible.x	= tx.width;
	default_ggi_mode.visible.y	= ty.width;
	default_ggi_mode.virt.x	= tx.width;
	default_ggi_mode.virt.y	= ty.width;
	default_ggi_mode.size.x		= 0;
	default_ggi_mode.size.y		= 0;
	default_ggi_mode.graphtype	= gmode;
	default_ggi_mode.text.x		= DCLKdiv;
	default_ggi_mode.text.y		= (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	if (dpy->defmode == NULL)
		dpy->defmode = &default_ggi_mode;
	else
		(*dpy->defmode) = default_ggi_mode;

	dpy->mode.request	= default_ggi_mode;
	dpy->mode.x		= tx;
	dpy->mode.y		= ty;
	dpy->mode.magnify.x	= magx;
	dpy->mode.magnify.y	= magy;
	dpy->mode.dclk		= dotclock;
	dpy->mode.clk.mul	= 1;
	dpy->mode.clk.div	= 1;
	dpy->mode.lclk.mul	= 1;
	dpy->mode.lclk.div	= 1;
	dpy->mode.chipset	= 0;
	
	default_kgi_mode = dpy->mode;

	chipset_setup_mode(&default_kgi_mode);
	chipset_setup_dpy(dpy, &(default_kgi_mode));

	return EOK;
}
   
static int chipset_init(struct kgi_display *dpy)
{
	int err;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (detect_VGA() == EOK) {
		if ((err = io_check_region(&vga_io)))
			return err;
		if ((err = mem_check_region(&vga_mem_graph)))
			return err;
#ifndef TEXT_A000
		if ((err = mem_check_region(&vga_mem_text)))
			return err;
#endif

		io_claim_region(&vga_io);
		mem_claim_region(&vga_mem_graph);
#ifndef TEXT_A000
		mem_claim_region(&vga_mem_text);
#endif

		text16_data_init();
		return chipset_get_reset_mode(dpy);
	}

	return -E(CHIPSET,FAILED);
}

static void chipset_reset(struct kgi_display *dpy)
{
	restore_state(&reset_state);
	if (dpy) {
		dpy->mode = default_kgi_mode;
		(*dpy->defmode) = default_ggi_mode;
	}
}

static void chipset_done(void)
{
	chipset_reset(NULL);

	io_free_region(&vga_io);
	mem_free_region(&vga_mem_graph);
#ifndef TEXT_A000
	mem_free_region(&vga_mem_text);
#endif
	
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == GT_TEXT16)
		? mode->request.text.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

static int chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
                              enum kgi_tm_cmd cmd)
{
	int bpp = 0;
	int textx, texty;
	textx = mode->request.text.x;
	texty = mode->request.text.y;

	switch (mode->request.graphtype) {
		case GT_8BIT:
		bpp = 8;	/* 1 Bpp */
		break; 	       
		case GT_1BIT:
		case GT_4BIT:
		bpp = 4;	/* 0.5 Bpp */
		break;	
		case GT_TEXT16:
		break;		/* Not used in textmode */
		default:
		DEBUG0("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET,FAILED);
	}

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (cmd) {
		case CMD_PROPOSE:
		if (mode->request.graphtype == GT_TEXT16) {

			mode->magnify.x = mode->magnify.y = 1;
			if ((textx < 8) || (9 < textx)) {
				DEBUG0("Invalid char width (%i)", textx);
				return -EINVAL;
			}

			mode->request.visible.x -=
				(mode->request.visible.x % textx);
			mode->request.visible.y -=
				(mode->request.visible.y % texty);

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
				return -EINVAL;

			if (XVIRTUAL == 0) 
				XVIRTUAL = fb_text_len / ((YVIRTUAL / texty) * 2) * textx;
			
			if (YVIRTUAL == 0)
				YVIRTUAL = fb_text_len / ((XVIRTUAL / textx) * 2) * texty;

			if ((2 * (XVIRTUAL / textx)
			     * (YVIRTUAL / texty)) > fb_text_len) {
				NOTICE("Not enough text memory");
				return -ENOMEM;
			}
		} else {	/* mode->request.graphtype == GT_TEXT16 */

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
				return -EINVAL;

			if (XVIRTUAL == 0)
				XVIRTUAL = fb_text_len / YVIRTUAL;

			if (YVIRTUAL == 0)
				YVIRTUAL = fb_graph_len / XVIRTUAL;

			mode->magnify.x = 1;
			mode->magnify.y = 1;

			while ((XVISIBLE * mode->magnify.x) < 640) 
				mode->magnify.x++;

			while ((YVISIBLE * mode->magnify.y) < 350) 
				mode->magnify.y++;
			
			/* Can't zoom more than this */
			if (mode->magnify.y > 2)
				return -EINVAL;
			
			/* Can't do more than doublescan */
			if (mode->magnify.x > 2)
				return -EINVAL;

			if ((textx < 8) || (9 < textx)) {
				DEBUG0("Invalid char width (%i)", textx);
				return -EINVAL;
			}
			if ((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize) {
				if ((XVISIBLE * YVISIBLE * bpp / 8) > chipinfo.ramsize) {
					DEBUG0("Not enough memory to do virtual %ix%i at %i bpp",
					       XVIRTUAL, YVIRTUAL, bpp);
					return -ENOMEM;
				} 
				XVIRTUAL = XVISIBLE;
				YVIRTUAL = YVISIBLE;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
		}

		mode->x.width = mode->request.visible.x * mode->magnify.x;
		mode->y.width = mode->request.visible.y * mode->magnify.y;

		return EOK;
		
		case CMD_LOWER:

		if (mode->dclk < chipinfo.dclk.min) {
			DEBUG1("DCLK too low (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk > chipinfo.dclk.max)
			? chipinfo.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_RAISE:

		if (mode->dclk > chipinfo.dclk.max) {
			DEBUG1("DCLK too high (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk < chipinfo.dclk.min)
			? chipinfo.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_CHECK:
		if (mode->lclk.mul != 1) {
			DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
			return -E(CHIPSET,FAILED);
		}
		switch (mode->lclk.div) {
			case 1:
			break;
			default:
			DEBUG1("Invalid CLK to LCLK divider %d", mode->lclk.div);
			return -E(CHIPSET,FAILED);
	    }
		if (mode->request.graphtype == GT_TEXT16) {
			mode->flags |= FL_CHIP_COLORTEXT;
			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
			    || (textx < 8) || (9 < textx) || (texty < 1)
			    || (texty > 32) || 
			    ((2 * (XVIRTUAL / textx) * 
			      (YVIRTUAL / texty)) > fb_text_len)
			    || (mode->dclk > chipinfo.dclk.max)
			    || (mode->dclk < chipinfo.dclk.min))
			{
				NOTICE1("HW check failed");
				return -E(CHIPSET,FAILED);
			}
		} else {
			switch (XVIRTUAL) {
				case 320:
				case 360:
				case 640:
				case 720:
				break;
				default:
				return -E(CHIPSET,FAILED);
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) > chipinfo.ramsize)
			    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
			    || (mode->clk.mul != 1) || (mode->clk.div != 1)
			    || (mode->dclk < chipinfo.dclk.min)
			    || (chipinfo.dclk.max < mode->dclk)
			    || (mode->magnify.x == 2 && mode->request.graphtype != GT_8BIT))
				return -E(CHIPSET,FAILED);
		}

#define TM(X) ((chipinfo.maxdots.x < mode->x.X) || (chipinfo.maxdots.y < mode->y.X))
		if (TM(width) || TM(blankstart) || TM(syncstart)
		    || TM(syncend) || TM(blankend) || TM(total)) {
			NOTICE1("Timing check failed");
			DEBUG1("Maxdots are (%d,%d)", chipinfo.maxdots.x,
				chipinfo.maxdots.y);
			return -E(CHIPSET,FAILED);
		}
#undef TM

		chipset_setup_mode(mode);
		return EOK;

		default:
		INTERNALERROR;
		return -E(CHIPSET,FAILED);
	}
	
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

/* This should probably be changed to an inline */
#define GGI_POKEBITS(VAL,MASK,REG) REG = ((REG & ~MASK) | (MASK & VAL))

static void chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv;
	int textx,texty;
	textx = mode->request.text.x;
	texty = mode->request.text.y;

	/* Set up bits per pixel and logical screen width */

	switch (mode->request.graphtype) {
		case GT_TEXT16:
		foo = 2 * mode->request.virt.x / textx;
		foo /= 4;	/* Due to the addressing mode used */
		break;
		case GT_1BIT:
		case GT_4BIT:
		foo = mode->request.virt.x >> 4;
		break;
		case GT_8BIT:
		foo = mode->request.virt.x >> 3;
		break;
		default:
		INTERNALERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH] = foo & 0xFF;

	if (mode->request.graphtype == GT_TEXT16) {
		setup.CRT[CRT_MODE] = 
		  CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC | 
		  CR17_CGA_BANKING | CR17_HGC_BANKING;

		setup.CRT[CRT_UNDERLINE] |= CR14_UNDERLINE_MASK;

		GGI_POKEBITS((SR02_PLANE0 | SR02_PLANE1),
			     SR02_PLANE_MASK, setup.SEQ[SEQ_WPLANE]);

		setup.GRC[GRC_MODECONTROL] |= GR5_ODD_EVEN;
		setup.GRC[GRC_MMAP_MODE] |= GR6_CHAIN_OE;
#if defined(TEXT_A000)
		GGI_POKEBITS(GR6_MAP_A0_64, GR6_MAP_MASK,
			     setup.GRC[GRC_MMAP_MODE]);
#else
		GGI_POKEBITS(GR6_MAP_B8_32, GR6_MAP_MASK,
			     setup.GRC[GRC_MMAP_MODE]);
#endif /* TEXT_A000 */

		setup.ATC[ATC_MODECONTROL] &= ~ 
		  (AR10_GRAPHICS | AR10_MONOMODE | 
		   0x10 /* reserved (0) */ | AR10_8BIT_MODE | AR10_ENB_BLINK);
		setup.ATC[ATC_MODECONTROL] |= (AR10_TOP_PAN | AR10_SELECT_B54);
		setup.ATC[ATC_BORDER] = 0;
		setup.ATC[ATC_HPAN] = 0;
		setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;

		switch ((DCLKdiv = textx)) {
			case 8:
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
			setup.ATC[ATC_MODECONTROL] &= ~AR10_9DOT_CHARS;
			break;
			case 9:
 			setup.SEQ[SEQ_CLOCK] &= ~SR01_8DOT_CHCLK;
			setup.ATC[ATC_MODECONTROL] |= AR10_9DOT_CHARS;
	 		break;
		 	default:
			INTERNALERROR;
			return;
		}
		GGI_POKEBITS((texty - 1), CR09_CHARHEIGHT_MASK,
			     setup.CRT[CRT_CHARHEIGHT]);

		GGI_POKEBITS(((textx == 9) ? 0x08 : 0x00), 
			     AR13_DATA_MASK, setup.ATC[ATC_HPAN]); 

	} else {	/* mode->request.graphtype == GT_TEXT16 */
		DCLKdiv = 8;

		setup.CRT[CRT_MODE] = 
		  CR17_CGA_BANKING | CR17_HGC_BANKING
		  | CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = default_state.ATC[foo];

		setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK;

			  GGI_POKEBITS(SR04_256K_ACCESS,
				       0x0f, setup.SEQ[SEQ_MEMMODE]);


		GGI_POKEBITS((GR6_GRAPHMODE | GR6_MAP_A0_64),
			     (GR6_GRAPHMODE | GR6_MAP_MASK),
			     setup.GRC[GRC_MMAP_MODE]);

		setup.CRT[CRT_HBLANKEND] |= CR03_IS_VGA;

		GGI_POKEBITS((mode->magnify.y - 1), CR09_CHARHEIGHT_MASK,
			     setup.CRT[CRT_CHARHEIGHT]);

		switch (mode->request.graphtype) {
			case GT_1BIT:
			case GT_4BIT:
			if (mode->request.graphtype == GT_1BIT) {
				setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK
				  /* Enable plane 0 for CPU */;
				GGI_POKEBITS(0x0e,
					     GR1_DATA_MASK,
					     setup.GRC[GRC_ENABLE]);
				GGI_POKEBITS(0x00,
					     GR4_PLANEMASK,
					     setup.GRC[GRC_READPLANE]);
				GGI_POKEBITS(GR5_WRITEMODE0,
					     0x03,
					     setup.GRC[GRC_MODECONTROL]); 
			} else {
			  /* Enable all planes for set/reset (color banking) */
			  GGI_POKEBITS(0x0f,
				       GR1_DATA_MASK,
				       setup.GRC[GRC_ENABLE]);
			  GGI_POKEBITS(GR5_WRITEMODE3,
				       0x03,
				       setup.GRC[GRC_MODECONTROL]); 
			}

			setup.SEQ[SEQ_MEMMODE] |= 
			  SR04_NO_ODDEVEN | SR04_256K_ACCESS;
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
			setup.CRT[CRT_MODE] |= CR17_BYTE_MODE;
			setup.ATC[ATC_MODECONTROL] |= AR10_GRAPHICS;

			break;

			case GT_8BIT:

			/* Note that this implies multiply.x==2 */
			setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;

			setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK;

			if (mode->request.virt.x == 320 &&
			    mode->request.virt.y == 200) {
				/* Planar */
			  GGI_POKEBITS((SR04_NO_ODDEVEN | SR04_CHAINED),
				       0x0f, setup.SEQ[SEQ_MEMMODE]); 
			  setup.CRT[CRT_UNDERLINE] |= CR14_DOUBLEWORDMODE;
			} else { 
			  /* Mode X */
			  GGI_POKEBITS(SR04_NO_ODDEVEN,
				       0x0f, setup.SEQ[SEQ_MEMMODE]);
			  setup.CRT[CRT_MODE] |= CR17_BYTE_MODE;
			}

			setup.ATC[ATC_MODECONTROL] &= ~ 
			  (AR10_GRAPHICS | AR10_MONOMODE | AR10_ENB_BLINK | 
			   0x10 /* reserved (0) */ | AR10_SELECT_B54);
			setup.ATC[ATC_MODECONTROL] |= 
			  (AR10_GRAPHICS | AR10_8BIT_MODE | AR10_TOP_PAN);
			setup.ATC[ATC_BORDER] = 0;
			setup.ATC[ATC_HPAN] = 0;
			setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;

			setup.GRC[GRC_MODECONTROL] |= GR5_8BIT_COLOR;
			setup.GRC[GRC_BITMASK] = 0xff;	/* Enable all bits */

			break;
			default:
			INTERNALERROR;
			return;
		}
	}
	
	/* Set up horizontal timing */
	
	foo = (mode->x.width / DCLKdiv) - 1;	/* Horizontal display end */
	setup.CRT[CRT_HDISPLAYEND] = foo & 0xFF;
	
	foo = (mode->x.blankstart / DCLKdiv);	/* Horizontal blank start */
	setup.CRT[CRT_HBLANKSTART] = foo & 0xFF;

	foo = (mode->x.syncstart / DCLKdiv);	/* Horizontal sync start  */
	setup.CRT[CRT_HSYNCSTART] = foo & 0xFF;

	foo = (mode->x.syncend / DCLKdiv);	/* Horizontal sync end    */
	GGI_POKEBITS(foo, CR05_HSE_MASK, setup.CRT[CRT_HSYNCEND]);

	foo = (mode->x.blankend / DCLKdiv);	/* Horizontal blank end   */
	GGI_POKEBITS(foo, CR03_HBE_MASK, setup.CRT[CRT_HBLANKEND]); 
	GGI_POKEBITS(((foo & 0x20) ? CR05_HBLANKEND_B5 : 0),
		     CR05_HBLANKEND_B5, setup.CRT[CRT_HSYNCEND]); 

	foo = (mode->x.total / DCLKdiv) - 5;	/* Horizontal total       */
	setup.CRT[CRT_HTOTAL] = foo & 0xFF;

	/* Set up vertical timing */

	foo = mode->y.width - 1;		/* Vertical display end   */
	setup.CRT[CRT_VDISPLAYEND] = foo & 0xFF;
	GGI_POKEBITS((((foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0) |
		      ((foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0)),
		     (CR07_VDISPLAYEND_B8 | CR07_VDISPLAYEND_B9), 
		     setup.CRT[CRT_OVERFLOW]); 
	
	foo = mode->y.blankstart - 1;		/* Vertical blank start   */
	setup.CRT[0x15] |= foo & 0xFF;
	GGI_POKEBITS(((foo & 0x100) ? CR07_VBLANKSTART_B8 : 0),		    
		      CR07_VBLANKSTART_B8, setup.CRT[CRT_OVERFLOW]); 
        GGI_POKEBITS(((foo & 0x200) ? CR09_VBLANKSTART_B9 : 0),
		      CR09_VBLANKSTART_B9, setup.CRT[CRT_CHARHEIGHT]); 

	foo = mode->y.syncstart;		/* Vertical sync start    */
	setup.CRT[CRT_VSYNCSTART] = foo & 0xFF;
	GGI_POKEBITS((((foo & 0x100) ? CR07_VSYNCSTART_B8 : 0) |
		      ((foo & 0x200) ? CR07_VSYNCSTART_B9 : 0)),
		     (CR07_VSYNCSTART_B8 | CR07_VSYNCSTART_B9), 
		     setup.CRT[CRT_OVERFLOW]); 

	foo = mode->y.syncend;			/* Vertical sync end      */
	GGI_POKEBITS(foo, CR11_VSYNCEND_MASK, setup.CRT[0x11]);

	foo = mode->y.blankend + 1;		/* Vertical blank end     */
	setup.CRT[CRT_VBLANKEND] = foo & 0xFF;

	foo = mode->y.total - 2;		/* Vertical total         */
	setup.CRT[CRT_VTOTAL] = foo & 0xFF;
	GGI_POKEBITS((((foo & 0x100) ? CR07_VTOTAL_B8 : 0) |
		      ((foo & 0x200) ? CR07_VTOTAL_B9 : 0)),
		      (CR07_VTOTAL_B8 | CR07_VTOTAL_B9), 
		      setup.CRT[CRT_OVERFLOW]); 

	GGI_POKEBITS((((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) |
		      ((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC)),
		     (MISC_NEG_HSYNC | MISC_NEG_VSYNC),
		     setup.MISC);

	/* Set splitline to 0 */

	setup.CRT[CRT_LINECOMPARE] = 0;
	GGI_POKEBITS(0, CR07_LINECOMPARE_B8, setup.CRT[CRT_OVERFLOW]);
	GGI_POKEBITS(0, CR09_LINECOMPARE_B9, setup.CRT[CRT_CHARHEIGHT]);

	/* Ok.  We'll do the actual hardware programming now. */
	
	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	kgim_clock_set_mode(dpy, mode);

	restore_state(&setup);

	chipset_setup_dpy(dpy, mode);

	NOTICE1("Mode set complete");
}

static int chipset_command(struct kgi_graphic *state,ggi_uint cmd,ggi_uint arg)
{
	ggi_coord  *coord;
	ggi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
		font = (struct kgi_font *)(state->io_buf);
		text16_set_font(dpy, font, 0, font->info->positions);
		return 0;
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.visible.y)
				return -E(CHIPSET,BOGUS);

			if (state->dev.org[0].y > state->dev.mode.visible.y - arg )
				kgi_set_origin(&(state->dev), 0, 
						state->dev.org[0].x,
						state->dev.mode.visible.y - arg );

			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (ggi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > state->dev.mode.virt.x - state->dev.mode.visible.x)
				return -E(CHIPSET,BOGUS);

			if (coord->y < 0 || dpy->mode.magnify.y<1 ||
			    coord->y > state->dev.mode.virt.y - state->dev.split/dpy->mode.magnify.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}

int kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	int err;

	if ((err = chipset_init(dpy)) != EOK) return err;
 
	strcpy(card->manufact, chipinfo.manufact);
	strcpy(card->model, chipinfo.model);
	card->modes = chipinfo.modes;
	card->flags = chipinfo.flags;
	card->maxdots = chipinfo.maxdots;

	card->chipset =& chipinfo;
	card->preset = default_kgi_mode;

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	chipset_done();
}

int  kgim_chipset_check_mode(struct kgi_display *dpy, struct kgi_mode *mode,
			     enum kgi_tm_cmd cmd)
{
	DEBUG1("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
	       mode->request.virt.y, mode->request.graphtype);
	return chipset_check_mode(dpy, mode, cmd);
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
  DEBUG1("Setting mode... (%dx%d %d)\n", mode->request.virt.x,
	 mode->request.virt.y, mode->request.graphtype);
	chipset_set_mode(dpy,mode);
}

int kgim_chipset_command(struct kgi_graphic *state, ggi_uint cmd, ggi_uint iobuf)
{
	return chipset_command(state, cmd, iobuf);
}
