/* --------------------------------------------------------------------------
**	MGA I/O routines
** --------------------------------------------------------------------------
**
**	Copyright (C)	1997		Jonas Borgstrom
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**	This code was derived from the following source of information:
**
**	Matrox mga2064w reference internet edition
**	XFree86 3.3.1 mga driver
**
**	$Log: mga_common_mmio.c,v $
**	Revision 1.2  1997/12/28 12:24:13  jonas_b
**	Applied Matthias patch and fixed some bugs.
**	
**	Revision 1.1  1997/11/22 11:25:56  jonas_b
**	Added a new driver which uses the "new linking scheme"
**	
**	Revision 1.1  1997/10/13 19:41:19  seeger
**	Initial checkin of Matrox Millenium driver by Jonas Borgstrom
**	
*/

#include <ggi/module.h>

#include <linux/pci.h>

#include <Matrox/mga_common_regs.h>
#include <Matrox/mga_common_mmio.h>
#include <Matrox/mga_common_text16.h>
#include <Matrox/mga_common.h>

/*
** VGA SEQuencer registers
*/
uint8 SEQ_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
	uint8 val;

	mem_out8(reg, MMIO_SEQI);
	val = mem_in8(MMIO_SEQD);
	DEBUG3("SEQ mem %.2x -> %.2x", reg, val);
	return val;
}

void SEQ_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
   
	DEBUG3("SEQ mem %.2x <- %.2x", reg, val);
	mem_out8(reg, MMIO_SEQI);
	mem_out8(val, MMIO_SEQD);
}

/*
** MISCellaneous output
*/
uint8 MISC_in8(struct kgi_display *dpy)
{
	GET_PRIVATE;
   
	register uint8 val = mem_in8(MMIO_MISCr);
	DEBUG3("MISC mem -> %.2x", val);
	return val;
}

void MISC_out8(struct kgi_display *dpy, uint8 val)
{
	GET_PRIVATE;
   
	mem_out8(val | MISC_COLOR_IO, MMIO_MISCw);
	DEBUG3("MISC mem <- %.2x", val | MISC_COLOR_IO);
}

/*
** VGA graphicx controller
*/
uint8 GRC_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
	uint8 val;
   
	mem_out8(reg, MMIO_GRCI);
	val = mem_in8(MMIO_GRCD);
	DEBUG3("GRC mem %.2x -> %.2x", reg, val);
	return val;
}

void GRC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
   
	DEBUG3("GRC mem %.2x <- %.2x", reg, val);
	mem_out8(reg, MMIO_GRCI);
	mem_out8(val, MMIO_GRCD);
}

/* 
** Matrox' extended CRTC registers 
*/
uint8 ECRT_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
   
	uint8 val;
 
	mem_out8(reg, MMIO_CRTXI);
	val = mem_in8(MMIO_CRTXD);
	DEBUG3("ECRT mem %.2x -> %.2x", reg, val);
	return val;
}

void ECRT_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
   
	DEBUG3("ECRT mem %.2x <- %.2x", reg, val);
	mem_out16((val << 8) | reg, MMIO_CRTXI);
}

/*
** VGA CRT registers
*/
uint8 CRT_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
   
	uint8 val;
 
	mem_out8(reg, MMIO_CRTI);
	val = mem_in8(MMIO_CRTD);
	DEBUG3("CRT mem %.2x -> %.2x", reg, val);
	return val;
}

void CRT_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
   
	DEBUG3("CRT mem %.2x <- %.2x", reg, val);
	mem_out8(reg, MMIO_CRTI);
	mem_out8(val, MMIO_CRTD);
}

/*
**	Clock I/O functions
*/
void CLK_out8(struct kgi_display *dpy, uint8 value, uint8 reg)
{
	GET_PRIVATE;
   
	ASSERT(reg == 0);
	mem_out8((mem_in8(MMIO_MISCr) & ~MISC_CLOCK_MASK) | 
		((value << 2) & MISC_CLOCK_MASK), MMIO_MISCw);
}
#if 0
uint8 CLK_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
   
	ASSERT(reg == 0);
	return (mem_in8(MMIO_MISCr) & MISC_CLOCK_MASK) >> 2;
}
#endif
/*
** Attribute controller
*/
#if 0
static uint8 ATC_in8(struct kgi_display *dpy, uint8 reg)
{
	GET_PRIVATE;
	uint8 val;

	mem_in8(MMIO_INS1r);	/* reset AFF and pas=0	*/
	mem_out8(reg, MMIO_ATCI);	/* set index	*/
	val = mem_in8(MMIO_ATCDr);	/* read value	*/
   
	mem_in8(MMIO_INS1r); /* reset AFF and pas=1, back to normal operation */
	mem_out8(ATCI_ENABLEDISPLAY, MMIO_ATCI);
	return val;
}
#endif
void ATC_out8(struct kgi_display *dpy, uint8 val, uint8 reg)
{
	GET_PRIVATE;
   
	mem_in8(MMIO_INS1r);	/* reset AFF and pas=0	*/
	mem_out8(reg, MMIO_ATCI);	/* set index	*/
	mem_out8(val, MMIO_ATCDw);	/* write value	*/
   
	mem_in8(MMIO_INS1r); /* reset AFF and pas=1, back to normal operation */
	mem_out8(ATCI_ENABLEDISPLAY, MMIO_ATCI);
}

void restore_ExtVga(struct kgi_display *dpy, uint8 *buf)
{
	GET_PRIVATE;
	register uint8 i;
   
	for (i = 0; i < NoExtVgaRegs; i++)
		mem_out16((*(buf++) << 8 ) | i, MMIO_CRTXI);
}

void restore_SEQ(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;

	for (i = 0; i < NoSEQRegs; i++)
		SEQ_out8(dpy, *(buf++), i);
}

void restore_CRT(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;
   
	CRT_out8(dpy, CRT_in8(dpy, CRT_VSYNCEND) & ~CR11_LOCKTIMING, CRT_VSYNCEND);
	CRT_out8(dpy, CRT_in8(dpy, CRT_HBLANKEND) | CR03_IS_VGA,CRT_HBLANKEND);
   
	for (i = 0; i < NoCRTRegs; i++) {

		switch (i) {

		case CRT_HBLANKEND:
			*buf |= CR03_IS_VGA;
			break;

		case CRT_VSYNCEND:
			*buf &= ~CR11_LOCKTIMING;
			break;

		default:
			break;
		}
		CRT_out8(dpy, *(buf++), i);
	}   
}

void restore_GRC(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;

	for (i = 0; i < NoGRCRegs; i++)
		GRC_out8(dpy, *(buf++), i);
}

void restore_ATC(struct kgi_display *dpy, uint8 *buf)
{
	register uint8 i;

	for (i = 0; i < NoATCRegs; i++)
		ATC_out8(dpy, *(buf++), i);
}
