/* --------------------------------------------------------------------------
**	text16 operators
** --------------------------------------------------------------------------
**
**	Copyright (C)	1997		Jochen Hoenicke
**			1997		Steffen Seeger
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
** ---------------------------------------------------------------------------
**
**	This code was derived from the following source of information:
**	[1]     Xfree 86 vga driver
**	[2]     Zen of Graphics Programming (Michael Abrash)
*/

#ifndef _S3_TEXT16_INC_
#define _S3_TEXT16_INC_

#if defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C765) || defined(S3_86C325)

	#define	MMIO_ONLY_TEXT
#endif

#if defined(S3_86C928)

	#define	NORMAL_TEXT
#endif

#define TEXTX           (dpy->mode.request.text.x)
#define TEXTY           (dpy->mode.request.text.y)
#define VIRTUALX        (dpy->fb.size.x)
#define VIRTUALY        (dpy->fb.size.y)
#define VISIBLEX        (dpy->mode.request.visible.x)
#define VISIBLEY        (dpy->mode.request.visible.y)

#ifdef MMIO_ONLY_TEXT

/*
**	MMIO only text mode operations
*/

#ifdef DEBUG_LEVEL
	/*	NOTE: the ASSERT lines should give you a ring if the framebuffer
	**	operations are used with inconsistent parameters. However, we
	**	rely on the OS to catch NULL pointer references!
	*/

	#define	FBSIZE	(fb->size.x*fb->size.y)
#endif


static void s3_text16_move(kgi_pixmap *fb, ggi_uint dst, ggi_uint src, ggi_uint cnt)
{
	register uint16 *pbuf = (void *) fb->pbuf;

	ASSERT((src + cnt) <= FBSIZE);
	ASSERT((dst + cnt) <= FBSIZE);
	ASSERT(src != dst);

	dst <<= 2;
	src <<= 2;

	if (dst < src) {

		while (cnt--) {

			mem_out16(mem_in16(pbuf + src), pbuf + dst);
			src += 4;
			dst += 4;
		}

	} else {

		src += (cnt - 1) << 2;
		dst += (cnt - 1) << 2;

		while (cnt--) {

			mem_out16(mem_in16(pbuf + src), pbuf + dst);
			src -= 4;
			dst -= 4;
		}
	}
}

static void s3_text16_set(kgi_pixmap *fb, ggi_uint dst, ggi_pixel val, ggi_uint cnt)
{
	register uint16 *pbuf = ((uint16 *) fb->pbuf) + (dst << 2);

	ASSERT((dst + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(val, pbuf);
		pbuf += 4;
	}
}

static void s3_text16_invert(kgi_pixmap *fb, ggi_uint start, ggi_uint end)
{
	register uint16 *pbuf = ((uint16 *)fb->pbuf) + (start << 2);
	register uint16 *pend = pbuf + (end << 2);

	ASSERT((start < FBSIZE) && (end < FBSIZE));
	ASSERT(start <= end);

	while (pbuf <= pend) {

		register uint16 h = mem_in16(pbuf);

		mem_out16((h & 0x88FF) | ((h<<4) & 0x7000) | ((h>>4) & 0x0700),
			pbuf);
		pbuf += 4;
	}
}

static void s3_text16_update(kgi_pixmap *fb, ggi_uint pos, ggi_pixel val)
{
	ASSERT(pos < FBSIZE);
	mem_out16(val, ((uint16 *) fb->pbuf) + (pos << 2));
}

static void s3_text16_flush(kgi_pixmap *fb)
{
}

static void s3_text16_copyfrom(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	register uint16 *src = ((uint16 *) fb->pbuf) + (start << 2);
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(*src, ((ushort *) buf)++);
		src += 4;
	}
}

static void s3_text16_copyto(kgi_pixmap *fb, void *buf, ggi_uint start, ggi_uint cnt)
{
	register uint16 *dst = ((uint16 *) fb->pbuf) + (start << 2);
	ASSERT((start + cnt) <= FBSIZE);

	while (cnt--) {

		mem_out16(*(((ushort *) buf)++), dst);
		dst += 4;
	}
}

/*	color mode attribute to pixel value mapping. We care of the foreground
**	and background colors. The rest is just best effort.
*/
static ggi_pixel s3_text16_atop(kgi_pixmap *fb, ggi_attr attr)
{
	register ggi_pixel
	result = (attr & ATTR_REVERSE)
			? (((attr & 0x0007) << 8) | ((attr & 0x0700) << 4))
			: (((attr & 0x0007) << 12) | (attr & 0x0700));

	if (attr & ATTR_BLINK) {

		result |= 0x8000;
	}

	if (attr & ATTR_BOLD) {

		result |= 0x0800;
	}

	return result;
}

static ggi_attr s3_text16_ptoa(kgi_pixmap *fb, ggi_pixel val)
{
	return	  ((val & 0x8000) ? ATTR_BLINK : 0)
		| ((val & 0x0800) ? ATTR_BOLD  : 0)
		| ((val & 0x7000) >> 12) | (val & 0x0F00);
}

/*	font postion to pixel mapping and vice versa. This cares about 256 
**	character fonts only (at least for now).
*/
static ggi_pixel s3_text16_ftop(kgi_pixmap *fb, ggi_uint pos)
{
	return pos & 0x00FF;
}

static ggi_uint s3_text16_ptof(kgi_pixmap *fb, ggi_pixel val)
{
	return val & 0x00FF;
}

static struct kgi_fb_ops s3_text16 = 
{
	s3_text16_move,
	s3_text16_set,
	s3_text16_invert,
	s3_text16_update,
	s3_text16_flush,
	s3_text16_copyfrom,
	s3_text16_copyto,

	s3_text16_atop,
	s3_text16_ptoa,
	s3_text16_ftop,
	s3_text16_ptof
};

/*
**	MMIO only text mode font setting
*/

static void do_set_font(struct kgi_display *dpy, int no, ggi_uint first,
	ggi_uint last, struct kgi_font *font)
{
	register uint32 *fbuf = (void *) dpy->fb.pbuf;
	register uchar *data = font->data + font->dy*first;

	if (font->data) {

		while (first <= last) {

			ggi_uint ggiindex = (first & 0xFF) + (no << 13);
			int i;

			for (i = 0; i < font->dy; i++, ggiindex += 0x100) {

				mem_out32((mem_in32(fbuf+ggiindex) & 0xFF00FFFF) |
					(data[i] << 16), fbuf+ggiindex);
			}
			for ( ; i < 32; i++, ggiindex += 0x100) {

				mem_out32(mem_in32(fbuf + ggiindex) & 0xFF00FFFF,
					fbuf + ggiindex);
			}

			first++;
			data += font->dy;
		}

        } else {

		for ( ; first <= last; first++) {

			ggi_uint ggiindex = first + (no << 13) + 32*0x100;
			int i;

			for (i = 33; --i; ggiindex -= 0x100) {

				mem_out32(mem_in32(fbuf + i) & 0xFF00FFFF,
					fbuf + i);
			}
		}
	}
}

#else	/* #ifdef MMIO_ONLY_TEXT */

static void do_set_font(struct kgi_display *dpy, int no, ggi_uint first,
	ggi_uint last, struct kgi_font *font)
{
	IS_CRITICAL;
	register GET_PRIVATE;
	uint8 *data, *fbuf = MEM_ACCEL;
	uint8 SR2, SR4, GR4, GR5, GR6, CR3A;
	register int i,j;

	if (font == NULL) {

		return;
	}
	fbuf += (no * 0x2000) + (first << 5);
	data = font->data + font->dy * first;

	/*	prepare hardware to access font plane
	*/
	ENTER_CRITICAL;

		/*	save register state
		*/
		SR2  = SEQ_in8(dpy, 0x02);
		SR4  = SEQ_in8(dpy, 0x04);
		GR4  = GRC_in8(dpy, 0x04);
		GR5  = GRC_in8(dpy, 0x05);
		GR6  = GRC_in8(dpy, 0x06);
		CR3A = CRT_in8(dpy, 0x3A);

		SEQ_out8(dpy, SR02_PLANE2, 0x02);    /* write to plane 2  */
		SEQ_out8(dpy, SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

		GRC_out8(dpy, 2, 0x04);	       /* read from plane 2 */
		GRC_out8(dpy, GR5_WRITEMODE0, 0x05);
		GRC_out8(dpy, GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);

		CRT_out8(dpy, CR3A | CR3A_HISPEEDTXT, 0x3A);
	
	LEAVE_CRITICAL;

	/* NOTE	no printk()'s are allowed in here unless klogd is active !
	*/

	if (data) {

		/*	set font if data is valid
		*/
		register int cnt = 32 - font->dy;

		for (i = first; i <= last; i++) {

			for (j = font->dy; j--; ) {

				mem_out8(*(data++), fbuf++);
			}
			for (j = cnt; j--; ) {

				mem_out8(0, fbuf++);
			}
		}

	} else {

		register uint32 *h = (void *) fbuf;

		/*	clear font
		*/
		for (i = first; i <= last; i++ ) {

			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);

			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
			mem_out32(0, h++);
		}
	}

	/*	restore hardware state
	*/
	ENTER_CRITICAL;

		CRT_out8(dpy, (CRT_in8(dpy, 0x0A) & ~CR0A_CURSOR_START_MASK) |
			((font->base + 1) & CR0A_CURSOR_START_MASK), 0x0A);
		CRT_out8(dpy, (CRT_in8(dpy, 0x0B) & ~CR0B_CURSOR_END_MASK) |
			((font->base + 2) & CR0B_CURSOR_END_MASK), 0x0B);

		SEQ_out8(dpy, SR2,  0x02);
		SEQ_out8(dpy, SR4,  0x04);
		GRC_out8(dpy, GR4,  0x04);
		GRC_out8(dpy, GR5,  0x05);
		GRC_out8(dpy, GR6,  0x06);
		CRT_out8(dpy, CR3A, 0x3A);

	LEAVE_CRITICAL;
}
#endif

static void s3_text_set_font(struct kgi_display *dpy, struct kgi_font *font,
			  ggi_uint from, ggi_uint to)
{
	do_set_font(dpy, 0, from, to, font);
	do_set_font(dpy, 1, from, to, font);
}

static void s3_text_set_origin(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	IS_CRITICAL;
	GET_PRIVATE;

	register ggi_uint xpix = x / TEXTX;
	register ggi_uint xdot = x % TEXTX;
	register ggi_uint ypix = y / TEXTY;
	register ggi_uint ydot = y % TEXTY;

	register ggi_uint offset = xpix + ypix*dpy->fb.size.x;
	private->origin = private->frame + offset;

	ENTER_CRITICAL;

		if (private->org_offs != offset) {

			CRT_out8(dpy, offset >> 8, CRT_STARTADDR_H);
			CRT_out8(dpy, offset,      CRT_STARTADDR_L);
			private->org_offs = offset;
		}

		if (private->org_ydot != ydot) {

			CRT_out8(dpy, ydot & CR08_ROWSCAN_MASK, CRT_PRESETROW);
			private->org_ydot = ydot;
		}

		if (private->org_xdot != xdot) {

			wait_display_off(dpy);

			if (dpy->mode.request.text.x == 9) {

				ATC_out8(dpy, (xdot == 9) ? 0x08 :
					(xdot-1) & AR13_DATA_MASK, ATC_HPAN);
			} else {

				ATC_out8(dpy, xdot & AR13_DATA_MASK, ATC_HPAN);
			}
			private->org_xdot = xdot;
		}

	LEAVE_CRITICAL;
}


static void s3_set_split(struct kgi_display *dpy, int line)
{
	GET_PRIVATE;
	register uchar foo;

	if (private->split == line) {

		return;	
	}
	private->split = line;

	CRT_out8(dpy, line & 0xFF, CRT_LINECOMPARE);

	foo = (line & 0x0100)
		? CRT_in8(dpy, 0x07) |  CR07_LINECOMPARE_B8
		: CRT_in8(dpy, 0x07) & ~CR07_LINECOMPARE_B8;
	CRT_out8(dpy, foo, 0x07);

	foo = (line & 0x0200)
		? CRT_in8(dpy, 0x09) |  CR09_LINECOMPARE_B9
		: CRT_in8(dpy, 0x09) & ~CR09_LINECOMPARE_B9;
	CRT_out8(dpy, foo, 0x09);

	foo = (line & 0x0400)
		? CRT_in8(dpy, 0x5E) |  CR5E_LINECOMPARE_B10
		: CRT_in8(dpy, 0x5E) & ~CR5E_LINECOMPARE_B10;
	CRT_out8(dpy, foo, 0x5E);
}

#define s3_set_frame NULL

static void s3_set_origin(struct kgi_display *dpy, ggi_uint dx, ggi_uint dy)
{
	int x = dx + dy*dpy->mode.request.virt.x;

	switch(dpy->mode.request.graphtype) {
	case GT_4BIT:	x >>= 1; break;
	case GT_15BIT:
	case GT_16BIT:	x <<= 1; break;
	case GT_32BIT:	x <<= 2; break;
	default:	break;
	}

#if defined(S3_86C765) || defined(S3_86C964) || defined(S3_86C968) || defined(S3_86C325)

	CRT_out8(dpy, (x >>  2) & 0xff, CRT_STARTADDR_L);    /*  sa2-sa9 */
	CRT_out8(dpy, (x >> 10) & 0xff, CRT_STARTADDR_H);    /* sa10-sa17 */
	CRT_out8(dpy, (x >> 18) & 0x0f, CRT_EXTCONTROL3 );

#elif defined(S3_86C928)

	x >>= 2;
	CRT_out8(dpy, x         & 0xff, CRT_STARTADDR_L);	/*  sa2-sa9  */
	CRT_out8(dpy, (x >>  8) & 0xff, CRT_STARTADDR_H);	/* sa10-sa17 */
	CRT_out8(dpy, (CRT_in8(dpy, CRT_MEMCONFIG0) & ~0x30) |
		((x >> 12) & 0x30), CRT_MEMCONFIG0);		/* sa18-sa19 */
	CRT_out8(dpy, (CRT_in8(dpy, CRT_SYSCONTROL2) & ~0x03) |
		((x >> 18) & 0x03), CRT_SYSCONTROL2);		/* sa20-sa21 */
#else
	#error no s3_set_origin, check this code!
#endif
}

/*
**	Cursor handling
*/

/*	hardware cursor
*/
static void s3_show_hw_text_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	register GET_PRIVATE;
	register ggi_uint pos = private->origin + 
			(x / TEXTX) + (y / TEXTY)*dpy->fb.size.x;

	pos = (pos < private->fb_size) ? pos : pos - private->fb_size;
	ASSERT(pos < private->fb_size);

	DEBUG2("cursor (%i,%i) = %i", x, y, pos);
	CRT_out8(dpy, pos >> 8, CRT_CURSORADDR_H);
	CRT_out8(dpy, pos,      CRT_CURSORADDR_L);
}

static void s3_hide_hw_text_cursor(struct kgi_display *dpy)
{
	CRT_out8(dpy, 0xFF, CRT_CURSORADDR_H);
	CRT_out8(dpy, 0xFF, CRT_CURSORADDR_L);
}

/*	software pointer and cursor
*/
#define	PTRAND	(private->ptr_and)
#define	PTRXOR	(private->ptr_xor)
#define	PTROLD	(private->ptr_old)

#define	CURAND	(private->cur_and)
#define	CURXOR	(private->cur_xor)
#define	CUROLD	(private->cur_old)

static void s3_show_sw_text_pointer(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	register GET_PRIVATE;
	register uint32 old = PTROLD;
	register uint32 cur = CUROLD;
	register uint32 new = private->origin +
		(x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	register uint16 *p = (uint16 *) dpy->fb.pbuf;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	new = (new < private->fb_size) ? new : new - private->fb_size;
	ASSERT(new < private->fb_size);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));
		#else
			#error implement software text mode pointer!
		#endif
	}

	#if defined(NORMAL_TEXT)
		p += new;
	#elif defined(MMIO_ONLY_TEXT)
		p += new << 2;
	#else
		#error implement software textmode pointer!
	#endif

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	PTROLD = new | 0x8000;
	mem_out16((mem_in16(p) & PTRAND) ^ PTRXOR, p);
}

static void s3_hide_sw_text_pointer(struct kgi_display *dpy)
{
	GET_PRIVATE;
	register uint32 old = PTROLD;
	register uint16 *p = (uint16 *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));

		#else
			#error implement sw_pointer_hide()
		#endif
	}

	PTROLD = 0;
}

#define	s3_undo_sw_text_pointer	s3_hide_sw_text_pointer

static void s3_show_sw_text_cursor(struct kgi_display *dpy, ggi_uint x, ggi_uint y)
{
	GET_PRIVATE;
	register uint32 old = CUROLD;
	register uint32 cur = PTROLD;
	register uint32 new = private->origin +
		(x / TEXTX)  +  (y / TEXTY) * VIRTUALX;
	register uint16 *p = (uint16 *) dpy->fb.pbuf;

	ASSERT((x < VISIBLEX) && (y < VISIBLEY));

	new = (new < private->fb_size) ? new : new - private->fb_size;
	ASSERT(new < private->fb_size);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));
		#else
			#error implement software text mode pointer!
		#endif
	}

	#if defined(NORMAL_TEXT)
		p += new;
	#elif defined(MMIO_ONLY_TEXT)
		p += new << 2;
	#else
		#error implement software textmode pointer!
	#endif

	if ((cur & 0x8000) && ((cur & 0x7FFF) == (new & 0x7FFF))) {

		new |= cur & 0xFFFF0000;

	} else {

		new |= mem_in16(p) << 16;
	}

	CUROLD = new | 0x8000;
	mem_out16((mem_in16(p) & CURAND) ^ CURXOR, p);
}

static void s3_hide_sw_text_cursor(struct kgi_display *dpy)
{
	GET_PRIVATE;
	register uint32 old = CUROLD;
	register uint16 *p = (uint16 *) dpy->fb.pbuf;
	ASSERT(p);

	if (old & 0x8000) {

		#if defined(MMIO_ONLY_TEXT)

			mem_out16(old >> 16, p + ((old & 0x7FFF) << 2));

		#elif defined(NORMAL_TEXT)

			mem_out16(old >> 16, p + (old & 0x7FFF));

		#else
			#error implement sw_pointer_hide()
		#endif
	}

	CUROLD = 0;
}

#define	s3_undo_sw_text_cursor	s3_hide_sw_text_cursor

#undef	PTRAND
#undef	PTRXOR
#undef	PTROLD

#undef	CURAND
#undef	CURXOR
#undef	CUROLD

#undef	TEXTX
#undef	TEXTY
#undef	VIRTUALX
#undef	VISIBLEX
#undef	VISIBLEY

#endif /* _S3_TEXT16_INC */
