/* ---------------------------------------------------------------------------
** Trio/ViRGE integrated PLL driver
** ---------------------------------------------------------------------------
**
** Copyright (C) 1997, The Linux-GGI project
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING.  If not, write to
** the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** 
** ---------------------------------------------------------------------------
** 
** $Log: s3idac.c,v $
** Revision 1.5  1997/12/14 21:55:59  becka
** uint/sint namespace fixed ... please test ...
**
** Revision 1.4  1997/12/03 15:11:59  seeger
** passing kgi_mode to pll_command()
**
** Revision 1.3  1997/10/14 03:33:31  taylorj
** Minor: more formatting cleanups
**
** Revision 1.2  1997/09/12 07:28:45  taylorj
** Minor: Reformatted source to be consistent with GGI coding style guidelines,
** modified all debugging output to use the DEBUGx macros, and added CVS
** revision logging to headers where needed.
**
** 
** ---------------------------------------------------------------------------
** This code was derived from the following source of information:
** [1] SDAC ramdac/clock synthesizer data book
*/

#include <ggi/maintainers.h>
#define	MAINTAINER	Jon_Taylor
#define	DRIVER_NAME	"S3 Trio/ViRGE integrated PLL clock driver"
#define	DRIVER_REV	"$Revision: 1.5 $"

#define	DEBUG_LEVEL	0

#include <ggi/module.h>

#include "S3/idac.h"
#include "pll.h"

static const struct pll_limits dclk_pll = 
{
	14318180,			/* Fref				*/
	{ 1, 1 },			/* Factor		 	*/
	{ 135000000, 270000000 },	/* FVCO range			*/
	{ 0, 3 },			/* Post-VCO divider range	*/
	{ 3, 129 },			/* Dividend range		*/
	{ 3, 33 }			/* Divisor range		*/
};

static struct kgi_clock clock =
{
	"S3 Incorporated",		/* Manufacturer		*/
	"Trio/ViRGE integrated DAC",	/* Model		*/
	GM_ALL,				/* Supported modes 	*/
	0,				/* Special capabilities	*/
	{ 0x7FFF, 0x7FFF },		/* Maximum resolution	*/
	CLK_TYPE_PROG,			/* Cock type		*/
	{{
		25000000, 110000000,	/* Clock range		*/
	}}
};

/* An SDAC command word is as follows:
**
**	[0:6]	s->mul - 2	
**	[8:12]	s->div - 2
**	[13:14]	s->p
*/

#define	INVALID_COMMAND	((ggi_uint) -1)

static ggi_uint pll_command(struct pll_state *s, struct kgi_mode *mode)
{
	ASSERT(s->mul - 2 < 128);
	ASSERT(s->div - 2 < 32);
	ASSERT(s->p < 4);

	return	(s->mul - 2) | ((s->div - 2) << 8) | (s->p << 13);
}

static void pll_set(struct kgi_display *dpy, ggi_uint cmd)
{
	/* Set clock select lines */
	CLK_out8(dpy, CLK0_PLL, 0);

	/* Set DCLK PLL parameters */
	CLK_out8(dpy, (cmd >> 8) & (CLK3_DCLK_N_MASK | CLK3_DCLK_R_MASK), 3);
	CLK_out8(dpy, cmd & CLK4_DCLK_M_MASK, 4);

	/* Load values */
#ifdef	DELAYED_LOAD
	CLK_out8(dpy, CLK6_LOAD_DLCK, 6);
#else
	CLK_out8(dpy, CLK6_LOAD_CLKS, 6);
	CLK_out8(dpy, CLK6_LOAD_DCLK, 6);
#endif
}

#include "pll.inc"
