/* --------------------------------------------------------------------------
**	S3 Virge Chipset Definitions
** --------------------------------------------------------------------------
**
**	Copyright (C)	1997		Stefan Mars
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
**
**	MAINTAINER	Stefan_Mars
**
**	$Log: virge.h,v $
**	Revision 1.7  1997/12/09 15:03:24  mars
**	Last update of s3_text.inc broke compilation of the virge
**	driver. Fixed.
**	
**	Revision 1.6  1997/11/25 21:03:44  mars
**	Corrected a spelling error
**	
**	Revision 1.5  1997/11/06 21:43:23  mars
**	Minor: Corrected problem with memory detection.
**	
**	Revision 1.4  1997/10/26 19:59:53  taylorj
**	Minor: Added more #defines for graphics modes
**	
**	Revision 1.3  1997/10/24 10:46:29  mars
**	*** empty log message ***
**	
**	Revision 1.2  1997/10/23 17:58:14  mars
**	*** empty log message ***
**	
**	Revision 1.1  1997/10/22 17:34:24  seeger
**	Initial ViRGE driver checkin
**	
**
** ---------------------------------------------------------------------------
*/
#ifndef	ViRGE_H
#define ViRGE_H

/*
** Including some additional files
*/
#include "IBM/vga.h"
#include "S3/accel.h"
#include "S3/idac.h"

/*
** MISC
*/
#undef	MISC_VIDEO_OFF
#define MISC_EXT_CLOCK			0x0C

/*
** CRT REGISTERS
*/

/*
** CRT2D - Device ID high register
*/
#define CR2D_ViRGE_HIGH_ID		0x56
#define CR2D_ViRGE_VX_HIGH_ID		0x88
#define CR2D_ViRGE_DX_GX_HIGH_ID	0x8A
#define CR2D_ViRGE_GX2_HIGH_ID		0x00
#define CR2D_ViRGE_MX_HIGH_ID		0x00

/*
** CRT2E - Device ID low register
*/
#define CR2E_ViRGE_LOW_ID		0x31
#define CR2E_ViRGE_VX_LOW_ID		0x3D
#define CR2E_ViRGE_DX_GX_LOW_ID		0x01
#define CR2E_ViRGE_GX2_LOW_ID		0x00
#define CR2E_ViRGE_MX_LOW_ID		0x00

/*
** CRT2F - Revision register
*/
#define CRT_REVISION			0x2F

/*
** CRT30 - Chip ID/rev register
*/
#define CRT_CHIPID			0x30
#define CR30_USE_CR2x			0xE1

/*
** CRT31 - Memory configuration
*/
#define CR31_HISPEEDTXT			0x40
#define CR31_ENHANCED			0x08
#define CR31_2PAGE_SCRN			0x02

/*
** CRT32 - Backward compatibility 1
*/


/*
** CRT33 - Backward Compatibility 2
*/
#define CR33_UNLOCK_VDE			0x02
#define CR33_LOCK_DAC			0x10
#define CR33_LOCK_PALETTE		0x40

/*
** CRT34 - Backward compatibility register 3
*/
#define	CRT_BACKCOMPAT3			0x34
#define	CR34_PCI_SNOOP			0x01
#define	CR34_PCI_ABORT			0x02
#define	CR34_PCI_RETRY			0x04
#define	CR34_ENB_FIFO			0x10

/*
** CRT36 - Configuration 1
*/
#define CRT_CONFIG1			0x36
#define CR36_VESALOCAL			0x01
#define CR36_PCI			0x02
#define CR36_BUSMASK			0x03
#define CR36_MEMSIZE_MASK		0xE0
#define CR36_1MB			0xC0
#define CR36_2MB			0x80
#define CR36_4MB			0x00
#define CR36_MEMMODE_MASK		0x0C
#define CR36_MEMMODE_1CYCLE		0x00
#define CR36_MEMMODE_2CYCLE		0x08
#define CR36_MEMMODE_SGRAM_SDRAM	0x40

/*
** CRT37 - Configuration 2
*/
#define CRT_CONFIG2			0x37

/*
** CRT38 - Lock 1
*/
#define CR38_MAGIC			0x48

/*
** CRT39 - Lock 2
*/
#define CR39_MAGIC			0xA0
#define CR39_CR36MAGIC			0xA5

/*
** CRT31 - Misc 1
*/
#define CR3A_REFRESHMASK		0x03
#define CR3A_PCIMASK			0x80
#define CR3A_NO4BPPMODE			0x10

/*
** CRT3B - Start display FIFO register
*/
#define CRT_FIFOSTART			0x3B

/*
** CRT3C - Interlace retrace start
*/
#define CRT_INTERLACESTART		0x3C

/*
** CRT42 - Mode control
*/
#define CR42_INTERLACED			0x20

/*
** CRT50 - Extended system control 1 - Undocumented
*/
#define	CRT_SYSCONTROL1			0x50
#define	CR50_WIDTHMASK			0xC1
#define	CR50_1024			0x00
#define	CR50_2048			0x00
#define CR50_640			0x40
#define	CR50_800			0x80
#define	CR50_1280			0xC0
#define	CR50_1152			0x01
#define	CR50_1600			0x81
#define	CR50_4BPP			0x00
#define	CR50_8BPP			0x00
#define	CR50_16BPP			0x10
#define	CR50_32BPP			0x30

/*
** CRT51 - Extended system control 2
*/
#define	CRT_SYSCONTROL2			0x51
#define	CR51_LOGICALWIDTH_B8		0x10
#define	CR51_LOGICALWIDTH_B9		0x20

/*
** CRT53 - Extended memory control 1
*/
#define	CRT_MEMCONTROL1			0x53
#define	CR53_MMIO_DISABLE		0x00
#define CR53_MMIO_RELOC_MMIO		0x08
#define	CR53_MMIO_TRIO64_LIKE		0x10
#define	CR53_MMIO_TRIO_AND_RELOC	0x18
#define	CR53_MMIO_A0_TO_B0		0x00
#define	CR53_MMIO_A0_TO_C0		0x20
#define	CR53_SWAPNIBBLES		0x40

/*
** CRT54 - Extended memory control 2
*/
#define	CRT_MEMCONTROL2			0x54
#define	CR54_M_MASK			0xF8
#define CR54_NO_BYTE_SWAP		0x00
#define	CR54_SWAP_BYTE_IN_WORD		0x01
#define	CR54_SWAP_BYTE_IN_DWORD		0x02
#define	CR54_SWAP_AS_PCI_OR_VL		0x03


/*
** CRT56 - External sync control 1
*/
#define CRT_EXTSYNC1			0x56
#define CR56_HSYNC_OFF			0x02
#define CR56_VSYNC_OFF			0x04

/*
** CRT58 - Linear address window control
*/
#define CR58_LAW_64KB			0x00
#define CR58_LAW_1MB			0x01
#define CR58_LAW_2MB			0x02
#define CR58_LAW_4MB			0x03
#define CR58_LAW_ENABLE			0x10
#define CR58_VL_T1_LATCH		0x40

/*
** CRT5D - Extended horizontal overflow
*/
#define CR5D_HDISPLAYEND_B8		0x02
#define CR5D_HBLANKSTART_B8		0x04
#define CR5D_HSYNCSTART_B8		0x10
#define CR5D_HSYNCEND_B5		0x20
#define CR5D_HBLANKEND_B6		0x40
#define CR5D_HTOTAL_B8			0x01
#define CR5D_FIFOSTART_B8		0x40

/*
** CRT5E - Extended vertical overflow
*/
#define CR5E_LINECOMPARE_B10		0x40
#define CR5E_VDISPLAYEND_B10		0x02
#define CR5E_VBLANKSTART_B10		0x04
#define CR5E_VSYNCSTART_B10		0x10
#define CR5E_VTOTAL_B10			0x01

/*
** CRT60 - Extended memory control 3 - Undocumented
*/
#define	CRT_MEMCONTROL3			0x60

/*
** CRT61 - Extended memory control 4
*/
#define	CRT_MEMCONTROL4			0x61
#define	CR61_NO_SWAP			0x00
#define	CR61_SWAP_BYTE_IN_WORD		0x20
#define	CR61_SWAP_BYTE_IN_DWORD		0x40

/*
** CRT66 - Extended misc control 1
*/
#define CR66_PCI_DISCONNECT		0x40

/*
** CRT67 - Extended misc control 2
*/
#define CR67_DAC_MODE0			0x00
#define CR67_DAC_MODE8			0x08
#define CR67_DAC_MODE9			0x30
#define CR67_DAC_MODEA			0x50


/*
** CRT68 - Configuration 3
*/
#define CRT_CONFIG3			0x68

/*
** CRT69 - Extended misc control 3
*/
#define CRT_EXTCONTROL3			0x69

/*
** SEQ REGISTERS
*/

/*
** SR08
*/
#define SR08_MAGIC			0x06

/*
** SR09
*/
#define	SEQ_IO_CTRL			0x09
#define	SR09_MMIO_ONLY			0x80

/*
** STREAMS REGISTERS
*/

/*
** MM8180 - Primary stream control
*/
#define STREAMS_PRIMARY_STREAMS_CONTROL			0x8180

/*
** MM8184 - Color/Chroma key control
*/
#define STREAMS_COLORCHROMA_KEY_CONTROL			0x8184

/*
** MM8190 - Secondary stream control
*/
#define STREAMS_SECONDARY_STREAMS_CONTROL		0x8190

/*
** MM8194 - Chroma key upper bound
*/
#define STREAMS_CHROMA_KEY_UPPER_BOUND			0x8194

/*
** MM8198 - Secondary stream stretch/filter constants
*/
#define STREAMS_SECONDARY_STRETCH_FILTER_CONSTANTS	0x8198

/*
** MM81A0 - Blend control
*/
#define STREAMS_BLEND_CONTROL				0x81A0

/*
** MM81C0 - Primary stream frame buffer address 0
*/
#define STREAMS_PRIMARY_STREAM_FB_ADDRESS0		0x81C0

/*
** MM81C4 - Primary stream frame buffer address 1
*/
#define STREAMS_PRIMARY_STREAM_FB_ADDRESS1		0x81C4

/*
** MM81C8 - Primary stream stride
*/
#define STREAMS_PRIMARY_STREAM_STRIDE			0x81C8

/*
** MM81CC - Double buffer/LPB support
*/
#define STREAMS_DOUBLE_BUFFER_LPB_SUPPORT		0x81CC

/*
** MM81D0 - Secondary stream frame buffer address 0
*/
#define STREAMS_SECONDARY_STREAM_FB_ADDRESS0		0x81D0

/*
** MM81D4 - Secondary stream frame buffer address 1
*/
#define STREAMS_SECONDARY_STREAM_FB_ADDRESS1		0x81D4

/*
** MM81D8 - Secondary stream stride
*/
#define STREAMS_SECONDARY_STREAM_STRIDE			0x81D8

/*
** MM81DC - Opaque overlay control
*/
#define STREAMS_OPAQUE_OVERLAY_CONTROL			0x81DC

/*
** MM81E0 - K1 vertical scale factor
*/
#define STREAMS_K1_VERTICAL_SCALE_FACTOR		0x81E0

/*
** MM81E4 - K2 vertical scale factor
*/
#define STREAMS_K2_VERTICAL_SCALE_FACTOR		0x81E4

/*
** MM81E8 - DDA vertical accumulator initial value
*/
#define STREAMS_DDA_INITIAL_VALUE			0x81E8

/*
** MM81EC - Streams FIFO control
*/
#define STREAMS_FIFO_CONTROL				0x81EC

/*
** MM81F0 - Primary stream window start coordinates
*/
#define STREAMS_PRIMARY_STREAM_WINDOW_START_COORD	0x81F0

/*
** MM81F4 - Primary stream window size
*/
#define STREAMS_PRIMARY_STREAM_WINDOW_SIZE		0x81F4

/*
** MM81F8 - Secondary stream window start coordinates
*/
#define STREAMS_SECONDARY_STREAM_WINDOW_START_COORD	0x81F8

/*
** MM81FC - Secondary stream window size
*/
#define STREAMS_SECONDARY_STREAM_WINDOW_SIZE		0x81FC


#endif /* ViRGE_H */
