/* ---------------------------------------------------------------------------
 *	PS/2 auxiliary pointing device driver
 * ---------------------------------------------------------------------------
 *
 *	Copyright (C) 1995-1997 by Steffen Seeger
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file COPYING.  If not, write to
 *	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * -------------------------------------------------------------------------
 * 
 * $Log: ps2aux.c,v $
 * Revision 1.8  1997/12/20 19:53:47  becka
 * Mouse driver fixed button order ...
 *
 * Revision 1.7  1997/12/14 21:56:11  becka
 * uint/sint namespace fixed ... please test ...
 *
 * Revision 1.6  1997/10/19 22:18:32  becka
 * Mouse numbering changed.
 *
 * Revision 1.5  1997/09/12 07:28:46  taylorj
 * Minor: Reformatted source to be consistent with GGI coding style guidelines,
 * modified all debugging output to use the DEBUGx macros, and added CVS
 * revision logging to headers where needed.
 *
 * 
 * ---------------------------------------------------------------------------
 * 
 */

#include <ggi/maintainers.h>
#define	MAINTAINER	Steffen_Seeger

#define DRIVER_NAME	"PS/2 Auxiliary mouse driver"
#define DRIVER_REV	"$Revision: 1.8 $"

#if defined(__alpha__) && !defined(CONFIG_PCI)

	#define AUX_IRQ	9	/* Jensen is odd indeed */

#else

	#define AUX_IRQ	12
#endif

#define	INIT_DEVICE			/* force initialization	*/

/* ===================================================================== *\
|* ----- there are no 'user servicable' parts below this line :-) ------ *|
\* ======================================================================*/

#include <linux/config.h>
#include <linux/module.h>

#include <ggi/kgi.h>

#include <asm/io.h>
#include <linux/random.h>

ggi_uint psaux_irq = AUX_IRQ;		/* IRQ to use		*/
ggi_uint psaux_focus = 0;			/* which focus to use	*/
extern uchar aux_device_present;
extern uchar psaux_read_mask;

static struct kgi_input psaux =
{
	NULL,				/* Registered devices	*/
	-1,				/* Display focused on	*/
	-1,				/* ID of this device	*/
	emPointer & ~emPtrAbsolute,	/* Events we generate	*/
	emPtrRelative | emPtrButton,	/* Events to report	*/
	"IBM",				/* Manufacturer		*/
	"PS/2 auxiliary pointer",	/* Model		*/
	NULL				/* ioctl function	*/
};

static ggi_sint state = 0;			/* decode state		*/
static ggi_uint buttons = 0;		/* button state		*/

/* psaux controller ports */
#define AUX_INPUT_PORT	0x60		/* Aux device output buffer	*/
#define AUX_OUTPUT_PORT	0x60		/* Aux device input buffer	*/
#define AUX_COMMAND	0x64		/* Aux device command buffer	*/
#define AUX_STATUS	0x64		/* Aux device status reg	*/

/* psaux controller status bits */
#define AUX_OBUF_FULL	0x21		/* outbuffer (from device) full	*/
#define AUX_IBUF_FULL	0x02		/* inbuffer (to device) full	*/

/* psaux controller commands */
#define AUX_CMD_WRITE	0x60		/* write to controller		*/
#define AUX_MAGIC_WRITE	0xd4		/* send psaux device data	*/

#define AUX_INTS_ON	0x47		/* enable controller interrupts	*/
#define AUX_INTS_OFF	0x65		/* disable controller interrupts*/

#define AUX_DISABLE	0xa7		/* disable psaux device		*/
#define AUX_ENABLE	0xa8		/* enable psaux			*/

/* psaux device commands */
#define AUX_SET_RES	0xe8		/* set resolution		*/
#define AUX_SET_SCALE11	0xe6		/* set 1:1 scaling		*/
#define AUX_SET_SCALE21	0xe7		/* set 2:1 scaling		*/
#define AUX_GET_SCALE	0xe9		/* get scaling factor		*/
#define AUX_SET_STREAM	0xea		/* set stream mode		*/
#define AUX_SET_SAMPLE	0xf3		/* set sample rate		*/
#define AUX_ENABLE_DEV	0xf4		/* enable psaux device		*/
#define AUX_DISABLE_DEV	0xf5		/* disable psaux device		*/
#define AUX_RESET	0xff		/* reset psaux device		*/

#define MAX_RETRIES	60		/* some operations take long time*/


static void psaux_interrupt(int cpl, void *dev_id, struct pt_regs * regs)
{
	static const char buttonmap[8] = { 0, 1, 2, 3, 4, 5, 6, 7 };
	static ggi_uint b;
	static ggi_sint dx, dy;
	ggi_event event;

	if ((inb(AUX_STATUS) & AUX_OBUF_FULL) != AUX_OBUF_FULL) {

		return;
	}

	switch (state) {

	case 0:	
		add_mouse_randomness(b = (uchar)inb(AUX_INPUT_PORT));
		state++;
		break;

	case 1:	
		add_mouse_randomness(dx = (uchar) inb(AUX_INPUT_PORT));
		if (b & 0x10) {

			dx |= (-1 & ~0xFF);
		}
		state++;
		break;

	case 2:	
		add_mouse_randomness(dy = (uchar) inb(AUX_INPUT_PORT));
		if (b & 0x20) {

			dy |= (-1 & ~0xFF);
		}
		state = 0;
		b = buttonmap[b & 7];

		DEBUG2("packet %.2x %i, %i", b, dx, dy);

		/* !!! do mouse acceleration here! */

		event.any.focus = psaux.focus;
		event.any.device = psaux.id;
		event.any.time = jiffies;

		/*	Change in position?
		*/
		if (dx || dy) 
		{

			event.pmove.size = sizeof(ggi_pmove_event);

			event.pmove.wheel = 0;
			event.pmove.dummy[0] = event.pmove.dummy[1] = 0;

			event.pmove.x =  dx;
			event.pmove.y = -dy;

			if (psaux.report & emPtrRelative) {

				NOTICE2("evPtrRelative");

				event.pmove.type = evPtrRelative;
				kgi_handle_input(&event);
			}
		}

		/*	Change in button state?
		*/
		if ((event.pbutton.button = buttons ^ b)) {

			event.pbutton.size = sizeof(ggi_pbutton_event);
			buttons = event.pbutton.state = b;
			event.pbutton.type =
				(event.pbutton.button & event.pbutton.state)
				? evPtrButtonPress : evPtrButtonRelease;

			if (psaux.report & (1 << event.pbutton.type)) {

				DEBUG2("%s", 
				       (event.pbutton.type == evPtrButtonPress)
				       ? "evPtrButtonPress" 
				       : "evPtrButtonRelease");
				kgi_handle_input(&event);
			}
		}
		break;
	}
}

static int psaux_poll_status(void)
{
	int retries = 0;

	while ((inb(AUX_STATUS) & 0x03) && (retries < MAX_RETRIES)) {

		if ((inb_p(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL) 
		{
			
			inb_p(AUX_INPUT_PORT);
		}
		current->state = TASK_INTERRUPTIBLE;
		current->timeout = jiffies + (5 * HZ + 99) / 100;
		schedule();
		retries++;
	}

	return retries != MAX_RETRIES;
}

static void psaux_write_dev(int val)
{
	psaux_poll_status();
	outb_p(AUX_MAGIC_WRITE, AUX_COMMAND);	/* Write magic cookie	*/
	psaux_poll_status();
	outb_p(val, AUX_OUTPUT_PORT);		/* Write data		*/
}

static void psaux_write_cmd(int val)
{
	psaux_poll_status();
	outb_p(AUX_CMD_WRITE, AUX_COMMAND);
	psaux_poll_status();
	outb_p(val, AUX_OUTPUT_PORT);
}

#if defined INIT_DEVICE
static int psaux_poll_status_nosleep(void)
{
	int retries = 0;

	while ((inb(AUX_STATUS) & 0x03) && retries < 1000000) {

		if ((inb_p(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL) {
			inb_p(AUX_INPUT_PORT);
		}
		retries++;
	}

	return retries != 1000000;
}

static int psaux_write_ack(int val)
{			/* Write and handle acknowledge	*/
	psaux_poll_status_nosleep();
	outb_p(AUX_MAGIC_WRITE, AUX_COMMAND);
	psaux_poll_status_nosleep();
	outb_p(val, AUX_OUTPUT_PORT);
	psaux_poll_status_nosleep();

	if ((inb(AUX_STATUS) & AUX_OBUF_FULL) == AUX_OBUF_FULL) {
		return (inb(AUX_INPUT_PORT));
	}

	return 0;
}
#endif

#ifdef MODULE
static
#endif
int psaux_init(void)
{
	if (aux_device_present != 0xaa) {

		DEBUG2("%s %s device not detected!", 
		       psaux.manufact, psaux.model);
		return -EIO;
	}

	if (request_irq(psaux_irq, psaux_interrupt, 0, psaux.model, NULL)) 
	{

		DEBUG2("Failed to claim IRQ %i.", psaux_irq);
		return -EIO;
	}

	DEBUG2("%s %s device detected.", psaux.manufact, psaux.model);
	psaux_read_mask = AUX_OBUF_FULL;

	outb_p(AUX_ENABLE, AUX_COMMAND);	/* Enable Aux		*/
	#ifdef INIT_DEVICE
		psaux_write_ack(AUX_SET_SAMPLE);
		psaux_write_ack(100);		/* 100 samples/sec	*/
		psaux_write_ack(AUX_SET_RES);
		psaux_write_ack(3);		/* 8 counts per mm	*/
		psaux_write_ack(AUX_SET_SCALE21);/* 2:1 scaling		*/
		psaux_poll_status_nosleep();
	#endif

	psaux_write_dev(AUX_ENABLE_DEV);
	psaux_write_cmd(AUX_INTS_ON);
	psaux_poll_status();
	state = 0;

	return kgi_register_input(psaux_focus, &psaux);
}

#ifdef MODULE
static int old_read_mask;

int init_module(void)
{
	old_read_mask = psaux_read_mask;
	return psaux_init();
}

void cleanup_module(void)
{
	psaux_write_cmd(AUX_INTS_OFF);
	psaux_poll_status();
	outb_p(AUX_DISABLE, AUX_COMMAND);
	psaux_poll_status();
	psaux_read_mask = old_read_mask;

	free_irq(psaux_irq, NULL);
	kgi_unregister_input(psaux_focus, &psaux);
	DEBUG1("%s %s driver unloaded.", psaux.manufact, psaux.model);
}
#endif
