/* --------------------------------------------------------------------------
**	IBM RGB526(DB) ramdac driver
** --------------------------------------------------------------------------
**
**	Copyright (C) 1997 Torgeir Veimo
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2, or (at your option)
**	any later version.
** 
**	You should have received a copy of the GNU General Public License
**	along with this program; see the file COPYING.  If not, write to
**	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
** 
** ---------------------------------------------------------------------------
** 
**	This code was derived from the following source of information:
**
**	[1]	XFree86 X server source code
**	[2]	IBM RGB526(DB) Databook (Preliminary).
*/

/* MAINTAINER	Torgeir Veimo, <Torgeir.Veimo@vertech.no> */

#warning This code is not working yet. Make sure you know what you are doing!

#include <ggi/maintainers.h>
#define MAINTAINER      Torgeir_Veimo
#define DRIVER_NAME     "IBM RGB256(DB) DAC driver"
#define DRIVER_REV      "$Revision: 1.4 $"
#define DEBUG_LEVEL     2

#define DAC_PRECISION   8	/* we can do 8 bit per RGB value */

#include <ggi/module.h>

#include "IBM/rgb52x.h"
#include "IBM/clut.inc"


#define SUPPORTED (GM_TEXT16 | GM_4BIT | GM_8BIT | \
	GM_15BIT | GM_16BIT | GM_24BIT | GM_32BIT)

static struct kgi_ramdac ramdac = {

	"IBM",			       	/* manufacturer		*/
	"RGB526(DB)",		       	/* model		*/
	SUPPORTED,			/* supported modes	*/
	0,			 	/* flags		*/
	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
  
	DAC_8BIT_PB,			/* device type		*/
	{	0,  32000000 },		/* normal mode		*/
	{	0,  32000000 },		/* 2xclk mode		*/
	
	{{	0,  32000000 },		/* 	GT_TEXT16	*/
	{	0,         0 },		/* 	GT_TEXT32 	*/
	{	0,  32000000 },		/* 	GT_1BIT		*/
	{	0,  32000000 },		/* 	GT_4BIT		*/
	{	0,  32000000 },		/* 	GT_8BIT		*/	
	{	0,  32000000 },		/* 	GT_15BIT	*/
	{	0,  32000000 },		/* 	GT_16BIT	*/
	{	0,  32000000 },		/* 	GT_24BIT	*/
	{	0,  32000000 }}		/* 	GT_32BIT	*/
};


/*	inline functions to access extended register set
*/

static inline void DAC_INT_out8(struct kgi_display *dpy, 
	uint8 value, uint16 ggiindex)
{
	DAC_out8(dpy, ggiindex, DAC_INDEX_LOW);
	DAC_out8(dpy, (ggiindex >> 8), DAC_INDEX_HIGH);
	DAC_out8(dpy, value, DAC_INDEX_DATA);
}

static inline uint8 DAC_INT_in8(struct kgi_display *dpy, uint16 ggiindex)
{
	DAC_out8(dpy, ggiindex, DAC_INDEX_LOW);
	DAC_out8(dpy, (ggiindex >> 8), DAC_INDEX_HIGH);
	return DAC_in8(dpy, DAC_INDEX_DATA);
}

/*	GGI interface
*/

int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	uint8 id, rev, tmp;

	/*	Probe
	*/
	id = DAC_INT_in8(dpy, DAC_ID);
	rev = DAC_INT_in8(dpy, DAC_REV);

	if (id != _RGB526_PROD_ID  ||  rev != _RGB526_REV_LEVEL) {

		NOTICE("No IBM RGB526 RAMDAC detected.");
		return -E(RAMDAC, UNKNOWN);
	}

	/*	These should not be here, since they are driver specific
	*/

	/*	Turn off autoincrement
	*/ 
	#warning Please check this!
	/*	This looks strange. convention for the DAC functions is
	**	DAC_in/out(dpy, value, register) if DAC_INDEX_CTRL is an
	**	register index, this use of the functions does not what you
	**	want, in short may be wrong! Please check -- seeger
	*/
	tmp = DAC_in8(dpy, DAC_INDEX_CTRL);
	DAC_out8(dpy, DAC_INDEX_CTRL, (tmp & ~1));

	/* Set Pixel Mask */
	DAC_out8(dpy, DAC_PIXEL_MASK, 0xFF);
  
	/* Set 8 bit palette operation */
	DAC_INT_out8(dpy, DAC_PAL, _RGB52X_6BIT_LINEAR_DISABLE | 
		_RGB52X_PALETTE_PARTITION(0));

	NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
	card->ramdac = &ramdac;

	return EOK;
}


void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}


int kgim_ramdac_check_mode (struct kgi_display *dpy, struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	/* ggi_uint newclock = mode->dclk; */
  
	switch (mode->request.graphtype) {

	case GT_4BIT:
		break;   
    
	case GT_8BIT:
		break;
  
	case GT_15BIT:
		break;

	case GT_16BIT:
		break;

	case GT_24BIT:
		break;

	case GT_32BIT:
		break;

	default:
		DEBUG2("couldn't handle graphtype %i", mode->request.graphtype);
		return -E(RAMDAC, MODE_CHECK);
	}

	switch (cmd) {

	case CMD_PROPOSE:
		mode->lclk.mul = mode->clk.mul = 1;
		mode->lclk.div = mode->clk.div = 1;
		return EOK;

	case CMD_LOWER:
	case CMD_RAISE:
		return EOK;

	case CMD_CHECK:
		if ((mode->lclk.mul != 1) || (mode->lclk.div != 1) ||
			(mode->clk.mul  != 1) || (mode->clk.div  != 1)) {
      
			return -E(RAMDAC,MODE_CHECK);
		}
		return EOK;

	default:
		INTERNAL_ERROR;
		return -E(RAMDAC, BOGUS);
	}

	return -E(RAMDAC, BOGUS);
}

void kgim_ramdac_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
    
	case GT_4BIT:
		break;
    
	case GT_8BIT:
		break;
    
	case GT_15BIT:
		break;
    
	case GT_16BIT:
		break;
    
	case GT_24BIT:
		break;
    
	case GT_32BIT:
		break;
    
	default:
		INTERNAL_ERROR;
	}

        dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, ggi_uint cmd, ggi_uint arg)
{
	ggi_clut clut;
  
	switch(cmd) { 
    
	case RAMDAC_SETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		dac_set_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;
  
	case RAMDAC_GETCLUT:
		clut.size = 256;
		clut.data = (ggi_color *)(graph->io_buf);
		graph->io_res = graph->io_buf;
		dac_get_clut(graph->dev.dpy, &clut, 0, 256);
		return 0;
	}

	return -E(RAMDAC, NOSUP_ALWAYS_CANT);  
}
