/* 
   Graphics library for GGI. Pixels.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not, write to the Free
  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "ggi-dl.h"
#include "color-magic.h"

/*********************************************
 * Let's pretend we have a 5:6:5 color cube! *
 *********************************************/

/* Return `5:6:6' mapping
 */
ggi_uint GGImapcolor(ggi_visual *vis,ggi_color col)
{
	return	((col.r>>(GGI_COLOR_PRECISION-5))<<11)|
		((col.g>>(GGI_COLOR_PRECISION-6))<<5)|
		(col.b>>(GGI_COLOR_PRECISION-5));
}

/* Map `5:6:6' pixel to color
 */
int GGIunmappixel(ggi_visual *vis,ggi_uint pixel,ggi_color *col)
{
	if (col==NULL) 
		return -1;

	col->b=(pixel & 0x1F) * MAGIC_5; 
	col->g=((pixel>>5)&0x3F) * MAGIC_6;
	col->r=((pixel>>11)&0x1F) * MAGIC_5;

	return EOK;	
}

/* Pack the colors into an array
 */
int GGIpackcolors(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	ushort *obuf=(ushort *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		*(obuf++)=(ushort)(ggiMapColor(vis,*(cols++)));

	return 0;
}	
	
/* Unpack into the ggi_color array the values of the pixels
 */
int GGIunpackpixels(ggi_visual *vis,void *outbuf,ggi_color *cols,int len)
{
	ushort *obuf=(ushort *)outbuf;
	int i;

	for (i=0;i<len;i++) 
		ggiUnmapPixel(vis,*(obuf++),(cols++));

	return 0;
}	
	
  
