/* 
   Graphics library for GGI.

   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu
   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ggi-dl.h"

/********************************/
/* draw/get/put a vertical line */
/********************************/
int GGIdrawvline(ggi_visual *vis,int x,int y,int height)
{
	uint8 *ptr;
	int w=VIS_FB_WIDTH(vis)>>1;
	uint8 shift = (x & 0x01) << 2;
	uint8 color=VIS_GC_FGCOLOR(vis) >> (shift^4);
	uint8 mask = 0x0f << shift;
	
	CHECKXYH(vis,x,y,height);

	ptr=(uint8 *)VIS_FB_LINEAR(vis)+y*((w+x)>>1);

	for(;height>0;height--,ptr+=w) 
		*ptr=color | (*ptr & mask); 
	
	return EOK;
}

int GGIputvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	uint8 *ptr,*pixels=(uint8 *)buffer;
	int w=VIS_FB_WIDTH(vis)>>1;
	uint8 shift = (x & 0x01) << 2;
	uint8 mask = 0x0f << shift;
	uint8 antishift = shift ^ 4;
	
	CHECKXYH(vis,x,y,height);

	ptr=(uint8 *)VIS_FB_LINEAR(vis)+y*((w+x)>>1);

	for(; height > 1; height-=2, ptr+=(w<<1)) {
		*ptr=(*pixels >> shift) | (*ptr & mask);
		*(ptr+w) = (*(pixels++) << antishift) | (*(ptr+w) & mask);
	}
	
	if (height) {
		*ptr=(*pixels >> shift) | (*ptr & mask);
	}
		
	return EOK;
}

int GGIgetvline(ggi_visual *vis,int x,int y,int height,void *buffer)
{
	uint8 *ptr,*pixels=(uint8 *)buffer;
	int w=VIS_FB_WIDTH(vis)>>1;
	uint8 shift = (x & 0x01) << 2;
	uint8 mask = 0x0f << shift;
	uint8 antishift = shift ^ 4;
	
	CHECKXYH(vis,x,y,height);

	/* Warning: unnecessary bit operations ahead! */
	for (; height > 1; height-=2, ptr+=w<<1) {
		*pixels = ((*ptr & mask) << shift) | ((*(ptr+w) & mask) >> antishift);
	}
	
	/* Here we can't lazily stick the extra pixel into the buffer, since
	 * it might be off the screen. */
	if (height) {
		*pixels = ((*ptr & mask) << shift);
	}
	
	return EOK;
}

