#ifndef _LIBGGI_DYNAMIC_H_
#define _LIBGGI_DYNAMIC_H_

/*
 * (c) 1997 Andreas Beck becka@sunserver1.rz.uni-duesseldorf.de
 * (c) 1997 Jason McMullan jmcc@ontv.com
 * (c) 1997 Steffen Seeger seeger@physik.tu-chemnitz.de
 * ------------------------------------------------------------------------
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU Library General Public License as 
 *	published by the Free Software Foundation; either version 2, or 
 *	(at your option) any later version.
 * 
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Library General Public License for more details.
 *
 *	You should have received a copy of the GNU Library General Public 
 *	License	along with this program; see the file COPYING.LGPL .  
 *	If not, write to the Free Software Foundation, 675 Mass Ave, 
 *	Cambridge, MA 02139, USA.
 *
 * ------------------------------------------------------------------------
 */


#include <sys/time.h>
#include <ggi/ggi.h>

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/* Do we need double here? */
typedef double ggi_float;

#ifndef _BUILDING_LIBGGI
typedef void *ggi_visual_t;
typedef void *ggi_lib_id;
#else
typedef struct ggi_visual *ggi_visual_t;
typedef struct ggi_dlhandle *ggi_lib_id;
#endif
typedef int (*ggi_lib_func)(ggi_visual_t vis,...);

#include <ggi/libggi_db.h>

/***********************************************************
 * Information that can be returned to user apps
 ***********************************************************/

/* Asynchronous [mulithreaded] or
 * synchronous [single-threaded] mode
 * default is sync 
 */
#define GGIFLAG_ASYNC	0x0001
#define GGIFLAG_SYNC	0x0000

typedef struct ggi_info_fb {
	int	bpp;	/* Bits per pixel */
	int	bpc;	/* Bits for color per pixel */
	int	bpa;	/* Bits for alpha per pixel */
	ggi_uint	width;	/* Width in pixels */
	ggi_uint	height;	/* Height in pixels */
	void   *linear;		/* Linear framebuffer */
	ggi_uint	linear_size;	/* Size of framebuffer in bytes */
} ggi_info_fb;

typedef struct ggi_info {
	ggi_uint		 flags;
	ggi_mode	*mode;
	ggi_info_fb	 fb;
	ggi_directbuffer *db;
	int		 select_fd;
} ggi_info;

/***************************************************************
 * libGGI function definitions 
 ***************************************************************/

/* Enter and leave the library
 */ 
int  ggiInit(void);
void ggiExit(void);
void ggiPanic(const char *format,...);

/* Open a new display - use display `NULL' for the current
 * virtual terminal
 */
ggi_visual_t ggiOpen(const char *display,...);
int          ggiClose(ggi_visual_t vis);
int	     ggiSetFocus(ggi_visual_t vis);
ggi_visual_t ggiGetFocus(void);

/* Get info 
 */
const ggi_info *ggiGetInfo(ggi_visual_t vis);
int ggiSetInfoFlags(ggi_visual_t vis,ggi_uint flags);

/* Library management
 */
ggi_lib_id   ggiLoadLibrary(ggi_visual_t vis,const char *name,const char *arg);
ggi_lib_func ggiGetLibraryFunc(ggi_visual_t vis,ggi_lib_id lib,ggi_uint funcid);

int	     ggiUnloadLibrary(ggi_visual_t vis,ggi_lib_id lib);
ggi_lib_id   ggiFindLibrary(ggi_visual_t vis,const char *);
int	     ggiRegisterExtension(int size,
			int (*load)(ggi_visual_t,const char *,const char *),
			int (*release)(ggi_visual_t,const char *));

/* Mode management
 */
#define GGI_AUTO (-1)	/* Can be given to use defaults. */

extern	ggi_mode ggiDefaultMode;

int	ggiSetMode(ggi_visual_t visual,ggi_mode *tm);
int	ggiGetMode(ggi_visual_t visual,ggi_mode *tm);
int	ggiCheckMode(ggi_visual_t visual,ggi_mode *tm);

int	ggiSetTextMode(ggi_visual_t visual,int cols,int rows,
					   int vcols,int vrows,
                                           int fontx,int fonty,
					   ggi_graphtype type);
int	ggiCheckTextMode(ggi_visual_t visual,int cols,int rows,
					     int vcols,int vrows,
                                             int fontx,int fonty,
					     ggi_graphtype type,
					     ggi_mode *suggested_mode,
					     ...);

int	ggiSetGraphMode(ggi_visual_t visual,int x,int y,
			int xv,int yv,ggi_graphtype type);
int	ggiCheckGraphMode(ggi_visual_t visual,int x,int y,
			  int xv,int yv,ggi_graphtype type,
			   ggi_mode *suggested_mode,...);

/*
 * Flush all pending operations to the display device
 */
int ggiFlush(ggi_visual_t visual);


/* WARNING - these are now depreciated! */
#define ggiGetFB(vis)  (ggiGetInfo(vis)->fb.linear)
#define ggiGetBPP(vis) (ggiGetInfo(vis)->fb.bpp)
#define ggiGetColorsPerPixel(vis) (1 << ggiGetInfo(vis)->fb.bpc)

/* Graphics context
 */
int ggiSetGCForeground(ggi_visual_t vis,ggi_uint  color);
int ggiGetGCForeground(ggi_visual_t vis,ggi_uint *color);
int ggiSetGCBackground(ggi_visual_t vis,ggi_uint  color);
int ggiGetGCBackground(ggi_visual_t vis,ggi_uint *color);

/* Color palette manipulation
 */
ggi_uint ggiMapColor(ggi_visual_t vis,ggi_color col);
int  ggiUnmapPixel(ggi_visual_t vis,ggi_uint pixel,ggi_color *col);

int  ggiPackColors(ggi_visual_t vis,void *buf,ggi_color *cols,int len);
int  ggiUnpackPixels(ggi_visual_t vis,void *buf,ggi_color *cols,int len);

int  ggiSetPaletteVec(ggi_visual_t vis,int s,int len,ggi_color *cmap);
int  ggiGetPaletteVec(ggi_visual_t vis,int s,int len,ggi_color *cmap);

/* Gamma map manipulation
 */
int  ggiGetGamma(ggi_visual_t vis,ggi_float *r,ggi_float *g,ggi_float *b);
int  ggiSetGamma(ggi_visual_t vis,ggi_float r,ggi_float g,ggi_float b);

int  ggiGetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);
int  ggiSetGammaMap(ggi_visual_t vis,int s,int len,ggi_color *gammamap);

/* Set origin and splitline features...
 */
int ggiSetOrigin(ggi_visual_t vis,int x,int y);
int ggiSetSplitline(ggi_visual_t vis,int y);

/* Graphics sprite manipulation
 */
void *ggiNewSprite(ggi_visual_t vis,int width,int height,void *image);
int ggiSetSprite(ggi_visual_t vis,void *pid,int v,int x,int y);
int ggiFreeSprite(ggi_visual_t vis,void *pid);

/* Generic drawing routines 
 */
int ggiPutc(ggi_visual_t vis,int x,int y,char c);
int ggiPuts(ggi_visual_t vis,int x,int y,const char *str);

int ggiFillscreen(ggi_visual_t vis);

int ggiDrawPixel(ggi_visual_t vis,int x,int y);
int ggiPutPixel(ggi_visual_t vis,int x,int y,ggi_uint col);
int ggiGetPixel(ggi_visual_t vis,int x,int y,ggi_uint *col);

int ggiDrawLine(ggi_visual_t vis,int x,int y,int xe,int ye);
int ggiDrawHLine(ggi_visual_t vis,int x,int y,int w);
int ggiPutHLine(ggi_visual_t vis,int x,int y,int w,void *buf);
int ggiGetHLine(ggi_visual_t vis,int x,int y,int w,void *buf);

int ggiDrawVLine(ggi_visual_t vis,int x,int y,int h);
int ggiPutVLine(ggi_visual_t vis,int x,int y,int h,void *buf);
int ggiGetVLine(ggi_visual_t vis,int x,int y,int h,void *buf);

int ggiDrawBox(ggi_visual_t vis,int x,int y,int w,int h);
int ggiPutBox(ggi_visual_t vis,int x,int y,int w,int h,void *buf);
int ggiGetBox(ggi_visual_t vis,int x,int y,int w,int h,void *buf);
int ggiCopyBox(ggi_visual_t vis,int x,int y,int w,int h,int nx,int ny);
int ggiCrossBlit(ggi_visual_t src,int sx,int sy,int sw,int sh,
		 ggi_visual_t dst,int dx,int dy,int dw,int dh);

int ggiDrawCircle(ggi_visual_t vis,int x,int y,int r);

#if (defined(__sgi) || defined(__osf__)) && !defined(__GNUC__)
#define inline
#endif

/* Event Handling */
ggi_event_mask ggiEventPoll(ggi_visual_t vis,ggi_event_mask mask,struct timeval *t);
int ggiEventRead(ggi_visual_t vis,ggi_event *ev,ggi_event_mask mask);

/* Convenience functions */
static inline int ggiKeyPoll(ggi_visual_t vis)
{
	struct timeval t={0,0};

	return (ggiEventPoll(vis, (ggi_event_mask) (emKeyPress|emKeyRepeat),&t) != 0);
}

static inline int ggiKeyRead(ggi_visual_t vis)
{
	ggi_event ev;

	/* Block until we get a key */
	ggiEventPoll(vis, (ggi_event_mask) (emKeyPress|emKeyRepeat),NULL);
	do {
		ggiEventRead(vis,&ev,emKeyboard);
	} while (ev.any.type != evKeyPress &&
		 ev.any.type != evKeyRepeat);

	return ev.key.sym;	
}

#define ggiKbhit(vis) 	ggiKeyPoll(vis)
#define ggiGetc(vis)	ggiKeyRead(vis)

#ifdef __sgi
#undef inline
#endif

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* _LIBGGI_DYNAMIC_H_ */
